from typing import Protocol, TypeVar, Optional, Any, Tuple, Union, Type, TypedDict, List
from abc import abstractmethod
from array_api_typing.typing_compat._api_typing import _NAMESPACE_ARRAY
from array_api_typing.typing_compat._array_typing import SetIndex

class AtResult(Protocol[_NAMESPACE_ARRAY]):
    @abstractmethod
    def __getitem__(self, idx: SetIndex, /) -> "AtResult[_NAMESPACE_ARRAY]":
        """
        Allow for the alternate syntax ``at(x)[start:stop:step]``.

        It looks prettier than ``at(x, slice(start, stop, step))``
        and feels more intuitive coming from the JAX documentation.
        """
        pass

    @abstractmethod
    def set(
        self,
        y: Union[_NAMESPACE_ARRAY, float, int, complex],
        /,
        copy: Optional[bool] = None,
    ) -> _NAMESPACE_ARRAY:
        pass

    @abstractmethod
    def add(
        self,
        y: Union[_NAMESPACE_ARRAY, float, int, complex],
        /,
        copy: Optional[bool] = None,
    ) -> _NAMESPACE_ARRAY:
        pass

    @abstractmethod
    def subtract(
        self,
        y: Union[_NAMESPACE_ARRAY, float, int, complex],
        /,
        copy: Optional[bool] = None,
    ) -> _NAMESPACE_ARRAY:
        pass

    @abstractmethod
    def multiply(
        self,
        y: Union[_NAMESPACE_ARRAY, float, int, complex],
        /,
        copy: Optional[bool] = None,
    ) -> _NAMESPACE_ARRAY:
        pass

    @abstractmethod
    def divide(
        self,
        y: Union[_NAMESPACE_ARRAY, float, int, complex],
        /,
        copy: Optional[bool] = None,
    ) -> _NAMESPACE_ARRAY:
        pass

    @abstractmethod
    def power(
        self,
        y: Union[_NAMESPACE_ARRAY, float, int, complex],
        /,
        copy: Optional[bool] = None,
    ) -> _NAMESPACE_ARRAY:
        pass

    @abstractmethod
    def min(
        self,
        y: Union[_NAMESPACE_ARRAY, float, int, complex],
        /,
        copy: Optional[bool] = None,
    ) -> _NAMESPACE_ARRAY:
        pass
    
    @abstractmethod
    def max(
        self,
        y: Union[_NAMESPACE_ARRAY, float, int, complex],
        /,
        copy: Optional[bool] = None,
    ) -> _NAMESPACE_ARRAY:
        pass


from typing import Any, Union, Callable, Mapping, Sequence
from array_api_typing.typing_compat import ArrayAPINamespace as CompatNamespace, ArrayAPIArray as CompatArray, ArrayAPIDType as CompatDType
import array_api_compat
import dataclasses

__all__ = [
    "get_abbreviate_array_function",
    "get_map_fn_over_arrays_function",
    "get_pad_dim_function",
]

def get_abbreviate_array_function(
    backend : CompatNamespace[CompatArray, Any, Any],
    default_integer_dtype : CompatDType,
    func_dtype_is_real_floating : Callable[[CompatDType], bool],
    func_dtype_is_real_integer : Callable[[CompatDType], bool],
    func_dtype_is_boolean : Callable[[CompatDType], bool],
):
    def abbreviate_array(array : CompatArray, try_cast_scalar : bool = True) -> Union[float, int, CompatArray]:
        """
        Abbreivates an array to a single element if possible.
        Or, if some dimensions are the same, abbreviates to a smaller array (but with the same number of dimensions).
        """
        abbr_array = array
        idx = backend.zeros(1, dtype=default_integer_dtype, device=array_api_compat.device(abbr_array))
        for dim_i in range(len(array.shape)):
            first_elem = backend.take(abbr_array, idx, axis=dim_i)
            if backend.all(abbr_array == first_elem):
                abbr_array = first_elem
            else:
                continue
        if try_cast_scalar:
            if all(i == 1 for i in abbr_array.shape):
                elem = abbr_array[tuple([0] * len(abbr_array.shape))]
                if func_dtype_is_real_floating(elem.dtype):
                    return float(elem)
                elif func_dtype_is_real_integer(elem.dtype):
                    return int(elem)
                elif func_dtype_is_boolean(elem.dtype):
                    return bool(elem)
                else:
                    raise ValueError(f"Abbreviated array element dtype must be a real floating or integer or boolean type, actual dtype: {elem.dtype}")
        else:
            return array
    return abbreviate_array

def get_map_fn_over_arrays_function(
    is_backendarray : Callable[[Any], bool],
):
    def map_fn_over_arrays(data : Any, func : Callable[[CompatArray], CompatArray]) -> Any:
        """
        Map a function to the data.
        """
        if is_backendarray(data):
            return func(data)
        elif isinstance(data, Mapping):
            ret = {k: map_fn_over_arrays(v, func) for k, v in data.items()}
            try:
                return type(data)(**ret)  # try to keep the same mapping type
            except:
                return ret
        elif isinstance(data, Sequence):
            ret = [map_fn_over_arrays(i, func) for i in data]
            try:
                return type(data)(ret)  # try to keep the same sequence type
            except:
                return ret
        elif dataclasses.is_dataclass(data):
            return type(data)(**map_fn_over_arrays(dataclasses.asdict(data), func))
        else:
            return data
    return map_fn_over_arrays

def get_pad_dim_function(
    backend : CompatNamespace[CompatArray, Any, Any],
):
    def pad_dim(x : CompatArray, dim : int, target_size : int, value : Union[float, int] = 0) -> CompatArray:
        size_at_target = x.shape[dim]
        if size_at_target > target_size:
            raise ValueError(f"Cannot pad dimension {dim} of size {size_at_target} to smaller target size {target_size}")
        if size_at_target == target_size:
            return x
        target_size = list(x.shape)
        target_size[dim] = target_size - size_at_target
        pad_value = backend.full(target_size, value, dtype=x.dtype, device=array_api_compat.device(x))
        return backend.concat([x, pad_value], axis=dim)
    return pad_dim


from typing import Protocol, TypeVar, Optional, Any, Tuple, Union, Type, TypedDict, List, Literal, Sequence
from abc import abstractmethod
from ._array_typing import *
from ._api_return_typing import *

_NAMESPACE_ARRAY = TypeVar("_NAMESPACE_ARRAY", bound=Array)
_NAMESPACE_DEVICE = TypeVar("_NAMESPACE_DEVICE", bound=Device)
_NAMESPACE_DTYPE = TypeVar("_NAMESPACE_DTYPE", bound=DType)
class ArrayAPILinalgNamespace(Protocol[_NAMESPACE_ARRAY, _NAMESPACE_DEVICE, _NAMESPACE_DTYPE]):
    """
    Linalg library API Stub
    https://github.com/data-apis/array-api/blob/main/src/array_api_stubs/_2024_12/linalg.py
    """
    @abstractmethod
    def cholesky(
        self, 
        x: _NAMESPACE_ARRAY, /, *, upper: bool = False
    ) -> _NAMESPACE_ARRAY:
        r"""
        Returns the lower (upper) Cholesky decomposition of a complex Hermitian or real symmetric positive-definite matrix ``x``.

        If ``x`` is real-valued, let :math:`\mathbb{K}` be the set of real numbers :math:`\mathbb{R}`, and, if ``x`` is complex-valued, let :math:`\mathbb{K}` be the set of complex numbers :math:`\mathbb{C}`.

        The lower **Cholesky decomposition** of a complex Hermitian or real symmetric positive-definite matrix :math:`x \in\ \mathbb{K}^{n \times n}` is defined as

        .. math::
        x = LL^{H} \qquad \text{L $\in\ \mathbb{K}^{n \times n}$}

        where :math:`L` is a lower triangular matrix and :math:`L^{H}` is the conjugate transpose when :math:`L` is complex-valued and the transpose when :math:`L` is real-valued.

        The upper Cholesky decomposition is defined similarly

        .. math::
        x = U^{H}U \qquad \text{U $\in\ \mathbb{K}^{n \times n}$}

        where :math:`U` is an upper triangular matrix.

        When ``x`` is a stack of matrices, the function must compute the Cholesky decomposition for each matrix in the stack.

        .. note::
        Whether an array library explicitly checks whether an input array is Hermitian or a symmetric positive-definite matrix (or a stack of matrices) is implementation-defined.

        Parameters
        ----------
        x: _NAMESPACE_ARRAY
            input array having shape ``(..., M, M)`` and whose innermost two dimensions form square complex Hermitian or real symmetric positive-definite matrices. Should have a floating-point data type.
        upper: bool
            If ``True``, the result must be the upper-triangular Cholesky factor :math:`U`. If ``False``, the result must be the lower-triangular Cholesky factor :math:`L`. Default: ``False``.

        Returns
        -------
        out: _NAMESPACE_ARRAY
            an array containing the Cholesky factors for each square matrix. If ``upper`` is ``False``, the returned array must contain lower-triangular matrices; otherwise, the returned array must contain upper-triangular matrices. The returned array must have a floating-point data type determined by :ref:`type-promotion` and must have the same shape as ``x``.

        Notes
        -----

        .. versionchanged:: 2022.12
        Added complex data type support.
        """

    @abstractmethod
    def cross(
        self,
        x1: _NAMESPACE_ARRAY, x2: _NAMESPACE_ARRAY, /, *, axis: int = -1
    ) -> _NAMESPACE_ARRAY:
        """
        Returns the cross product of 3-element vectors.

        If ``x1`` and/or ``x2`` are multi-dimensional arrays (i.e., the broadcasted result has a rank greater than ``1``), then the cross-product of each pair of corresponding 3-element vectors is independently computed.

        Parameters
        ----------
        x1: _NAMESPACE_ARRAY
            first input array. Must have a numeric data type. The size of the axis over which the cross product is to be computed must be equal to 3.
        x2: _NAMESPACE_ARRAY
            second input array. Must be broadcast compatible with ``x1`` along all axes other than the axis along which the cross-product is computed (see :ref:`broadcasting`). The size of the axis over which the cross product is to be computed must be equal to 3. Must have a numeric data type.

            .. note::
            The compute axis (dimension) must not be broadcasted.

        axis: int
            the axis (dimension) of ``x1`` and ``x2`` containing the vectors for which to compute the cross product. Should be an integer on the interval ``[-N, -1]``, where ``N`` is ``min(x1.ndim, x2.ndim)``. The function must determine the axis along which to compute the cross product by counting backward from the last dimension (where ``-1`` refers to the last dimension). By default, the function must compute the cross product over the last axis. Default: ``-1``.

        Returns
        -------
        out: _NAMESPACE_ARRAY
            an array containing the cross products. The returned array must have a data type determined by :ref:`type-promotion`.


        Notes
        -----

        **Raises**

        -   if the size of the axis over which to compute the cross product is not equal to ``3`` (before broadcasting) for both ``x1`` and ``x2``.

        .. versionchanged:: 2022.12
        Added support for broadcasting.

        .. versionchanged:: 2022.12
        Added complex data type support.

        .. versionchanged:: 2023.12
        Restricted broadcasting to only non-compute axes and required that ``axis`` be a negative integer.
        """

    @abstractmethod
    def det(
        self,
        x: _NAMESPACE_ARRAY, /
    ) -> _NAMESPACE_ARRAY:
        """
        Returns the determinant of a square matrix (or a stack of square matrices) ``x``.

        Parameters
        ----------
        x: _NAMESPACE_ARRAY
            input array having shape ``(..., M, M)`` and whose innermost two dimensions form square matrices. Should have a floating-point data type.

        Returns
        -------
        out: _NAMESPACE_ARRAY
            if ``x`` is a two-dimensional array, a zero-dimensional array containing the determinant; otherwise, a non-zero dimensional array containing the determinant for each square matrix. The returned array must have the same data type as ``x``.

        Notes
        -----

        .. versionchanged:: 2022.12
        Added complex data type support.
        """

    @abstractmethod
    def diagonal(
        self,
        x: _NAMESPACE_ARRAY, /, *, offset: int = 0
    ) -> _NAMESPACE_ARRAY:
        """
        Returns the specified diagonals of a matrix (or a stack of matrices) ``x``.

        Parameters
        ----------
        x: _NAMESPACE_ARRAY
            input array having shape ``(..., M, N)`` and whose innermost two dimensions form ``MxN`` matrices.
        offset: int
            offset specifying the off-diagonal relative to the main diagonal.

            - ``offset = 0``: the main diagonal.
            - ``offset > 0``: off-diagonal above the main diagonal.
            - ``offset < 0``: off-diagonal below the main diagonal.

            Default: `0`.

        Returns
        -------
        out: _NAMESPACE_ARRAY
            an array containing the diagonals and whose shape is determined by removing the last two dimensions and appending a dimension equal to the size of the resulting diagonals. The returned array must have the same data type as ``x``.
        """

    @abstractmethod
    def eigh(
        self,
        x: _NAMESPACE_ARRAY, /
    ) -> Tuple[_NAMESPACE_ARRAY, _NAMESPACE_ARRAY]:
        r"""
        Returns an eigenvalue decomposition of a complex Hermitian or real symmetric matrix (or a stack of matrices) ``x``.

        If ``x`` is real-valued, let :math:`\mathbb{K}` be the set of real numbers :math:`\mathbb{R}`, and, if ``x`` is complex-valued, let :math:`\mathbb{K}` be the set of complex numbers :math:`\mathbb{C}`.

        The **eigenvalue decomposition** of a complex Hermitian or real symmetric matrix :math:`x \in\ \mathbb{K}^{n \times n}` is defined as

        .. math::
        x = Q \Lambda Q^H

        with :math:`Q \in \mathbb{K}^{n \times n}` and :math:`\Lambda \in \mathbb{R}^n` and where :math:`Q^H` is the conjugate transpose when :math:`Q` is complex and the transpose when :math:`Q` is real-valued and :math:`\Lambda` is a diagonal matrix whose diagonal elements are the corresponding eigenvalues. When ``x`` is real-valued, :math:`Q` is orthogonal, and, when ``x`` is complex, :math:`Q` is unitary.

        .. note::
        The eigenvalues of a complex Hermitian or real symmetric matrix are always real.

        .. warning::
        The eigenvectors of a symmetric matrix are not unique and are not continuous with respect to ``x``. Because eigenvectors are not unique, different hardware and software may compute different eigenvectors.

        Non-uniqueness stems from the fact that multiplying an eigenvector by :math:`-1` when ``x`` is real-valued and by :math:`e^{\phi j}` (:math:`\phi \in \mathbb{R}`) when ``x`` is complex produces another set of valid eigenvectors.

        .. note::
        Whether an array library explicitly checks whether an input array is Hermitian or a symmetric matrix (or a stack of matrices) is implementation-defined.

        .. note::
        The function ``eig`` will be added in a future version of the specification.

        Parameters
        ----------
        x: _NAMESPACE_ARRAY
            input array having shape ``(..., M, M)`` and whose innermost two dimensions form square matrices. Should have a floating-point data type.

        Returns
        -------
        out: Tuple[array, array]
            a namedtuple (``eigenvalues``, ``eigenvectors``) whose

            -   first element must have the field name ``eigenvalues`` (corresponding to :math:`\operatorname{diag}\Lambda` above) and must be an array consisting of computed eigenvalues. The array containing the eigenvalues must have shape ``(..., M)`` and must have a real-valued floating-point data type whose precision matches the precision of ``x`` (e.g., if ``x`` is ``complex128``, then ``eigenvalues`` must be ``float64``).
            -   second element have have the field name ``eigenvectors`` (corresponding to :math:`Q` above) and must be an array where the columns of the inner most matrices contain the computed eigenvectors. These matrices must be orthogonal. The array containing the eigenvectors must have shape ``(..., M, M)`` and must have the same data type as ``x``.

        Notes
        -----

        .. note::
        Eigenvalue sort order is left unspecified and is thus implementation-dependent.

        .. versionchanged:: 2022.12
        Added complex data type support.
        """

    @abstractmethod
    def eigvalsh(
        self,
        x: _NAMESPACE_ARRAY, /
    ) -> _NAMESPACE_ARRAY:
        r"""
        Returns the eigenvalues of a complex Hermitian or real symmetric matrix (or a stack of matrices) ``x``.

        If ``x`` is real-valued, let :math:`\mathbb{K}` be the set of real numbers :math:`\mathbb{R}`, and, if ``x`` is complex-valued, let :math:`\mathbb{K}` be the set of complex numbers :math:`\mathbb{C}`.

        The **eigenvalues** of a complex Hermitian or real symmetric matrix :math:`x \in\ \mathbb{K}^{n \times n}` are defined as the roots (counted with multiplicity) of the polynomial :math:`p` of degree :math:`n` given by

        .. math::
        p(\lambda) = \operatorname{det}(x - \lambda I_n)

        where :math:`\lambda \in \mathbb{R}` and where :math:`I_n` is the *n*-dimensional identity matrix.

        .. note:;
        The eigenvalues of a complex Hermitian or real symmetric matrix are always real.

        .. note::
        Whether an array library explicitly checks whether an input array is Hermitian or a symmetric matrix (or a stack of matrices) is implementation-defined.

        .. note::
        The function ``eigvals`` will be added in a future version of the specification.

        Parameters
        ----------
        x: _NAMESPACE_ARRAY
            input array having shape ``(..., M, M)`` and whose innermost two dimensions form square matrices. Should have a floating-point data type.

        Returns
        -------
        out: _NAMESPACE_ARRAY
            an array containing the computed eigenvalues. The returned array must have shape ``(..., M)`` and have a real-valued floating-point data type whose precision matches the precision of ``x`` (e.g., if ``x`` is ``complex128``, then must have a ``float64`` data type).

        Notes
        -----

        .. note::
        Eigenvalue sort order is left unspecified and is thus implementation-dependent.

        .. versionchanged:: 2022.12
        Added complex data type support.
        """

    @abstractmethod
    def inv(
        self,
        x: _NAMESPACE_ARRAY, /
    ) -> _NAMESPACE_ARRAY:
        r"""
        Returns the multiplicative inverse of a square matrix (or a stack of square matrices) ``x``.

        If ``x`` is real-valued, let :math:`\mathbb{K}` be the set of real numbers :math:`\mathbb{R}`, and, if ``x`` is complex-valued, let :math:`\mathbb{K}` be the set of complex numbers :math:`\mathbb{C}`.

        The **inverse matrix** :math:`x^{-1} \in\ \mathbb{K}^{n \times n}` of a square matrix :math:`x \in\ \mathbb{K}^{n \times n}` is defined as

        .. math::
        x^{-1}x = xx^{-1} = I_n

        where :math:`I_n` is the *n*-dimensional identity matrix.

        The inverse matrix exists if and only if ``x`` is invertible. When ``x`` is invertible, the inverse is unique.

        When ``x`` is a stack of matrices, the function must compute the inverse for each matrix in the stack.

        Parameters
        ----------
        x: _NAMESPACE_ARRAY
            input array having shape ``(..., M, M)`` and whose innermost two dimensions form square matrices. Should have a floating-point data type.

        Returns
        -------
        out: _NAMESPACE_ARRAY
            an array containing the multiplicative inverses. The returned array must have a floating-point data type determined by :ref:`type-promotion` and must have the same shape as ``x``.

        Notes
        -----

        .. versionchanged:: 2022.12
        Added complex data type support.
        """

    @abstractmethod
    def matmul(
        self,
        x1: _NAMESPACE_ARRAY, x2: _NAMESPACE_ARRAY, /
    ) -> _NAMESPACE_ARRAY:
        """Alias for :func:`~array_api.matmul`."""

    @abstractmethod
    def matrix_norm(
        self,
        x: _NAMESPACE_ARRAY,
        /,
        *,
        keepdims: bool = False,
        ord: Optional[Union[int, float, Literal["fro", "nuc"]]] = "fro",
    ) -> _NAMESPACE_ARRAY:
        """
        Computes the matrix norm of a matrix (or a stack of matrices) ``x``.

        Parameters
        ----------
        x: _NAMESPACE_ARRAY
            input array having shape ``(..., M, N)`` and whose innermost two dimensions form ``MxN`` matrices. Should have a floating-point data type.
        keepdims: bool
            If ``True``, the last two axes (dimensions) must be included in the result as singleton dimensions, and, accordingly, the result must be compatible with the input array (see :ref:`broadcasting`). Otherwise, if ``False``, the last two axes (dimensions) must not be included in the result. Default: ``False``.
        ord: Optional[Union[int, float, Literal[inf, -inf, 'fro', 'nuc']]]
            order of the norm. The following mathematical norms must be supported:

            +------------------+---------------------------------+
            | ord              | description                     |
            +==================+=================================+
            | 'fro'            | Frobenius norm                  |
            +------------------+---------------------------------+
            | 'nuc'            | nuclear norm                    |
            +------------------+---------------------------------+
            | 1                | max(sum(abs(x), axis=0))        |
            +------------------+---------------------------------+
            | 2                | largest singular value          |
            +------------------+---------------------------------+
            | inf              | max(sum(abs(x), axis=1))        |
            +------------------+---------------------------------+

            The following non-mathematical "norms" must be supported:

            +------------------+---------------------------------+
            | ord              | description                     |
            +==================+=================================+
            | -1               | min(sum(abs(x), axis=0))        |
            +------------------+---------------------------------+
            | -2               | smallest singular value         |
            +------------------+---------------------------------+
            | -inf             | min(sum(abs(x), axis=1))        |
            +------------------+---------------------------------+

            If ``ord=1``, the norm corresponds to the induced matrix norm where ``p=1`` (i.e., the maximum absolute value column sum).

            If ``ord=2``, the norm corresponds to the induced matrix norm where ``p=inf`` (i.e., the maximum absolute value row sum).

            If ``ord=inf``, the norm corresponds to the induced matrix norm where ``p=2`` (i.e., the largest singular value).

            Default: ``'fro'``.

        Returns
        -------
        out: _NAMESPACE_ARRAY
            an array containing the norms for each ``MxN`` matrix. If ``keepdims`` is ``False``, the returned array must have a rank which is two less than the rank of ``x``. If ``x`` has a real-valued data type, the returned array must have a real-valued floating-point data type determined by :ref:`type-promotion`. If ``x`` has a complex-valued data type, the returned array must have a real-valued floating-point data type whose precision matches the precision of ``x`` (e.g., if ``x`` is ``complex128``, then the returned array must have a ``float64`` data type).

        Notes
        -----

        .. versionchanged:: 2022.12
        Added complex data type support.
        """

    @abstractmethod
    def matrix_power(
        self,
        x: _NAMESPACE_ARRAY, n: int, /
    ) -> _NAMESPACE_ARRAY:
        """
        Raises a square matrix (or a stack of square matrices) ``x`` to an integer power ``n``.

        Parameters
        ----------
        x: _NAMESPACE_ARRAY
            input array having shape ``(..., M, M)`` and whose innermost two dimensions form square matrices. Should have a floating-point data type.
        n: int
            integer exponent.

        Returns
        -------
        out: _NAMESPACE_ARRAY
            if ``n`` is equal to zero, an array containing the identity matrix for each square matrix. If ``n`` is less than zero, an array containing the inverse of each square matrix raised to the absolute value of ``n``, provided that each square matrix is invertible. If ``n`` is greater than zero, an array containing the result of raising each square matrix to the power ``n``. The returned array must have the same shape as ``x`` and a floating-point data type determined by :ref:`type-promotion`.

        Notes
        -----

        .. versionchanged:: 2022.12
        Added complex data type support.
        """

    @abstractmethod
    def matrix_rank(
        self,
        x: _NAMESPACE_ARRAY, /, *, rtol: Optional[Union[float, _NAMESPACE_ARRAY]] = None
    ) -> _NAMESPACE_ARRAY:
        """
        Returns the rank (i.e., number of non-zero singular values) of a matrix (or a stack of matrices).

        When ``x`` is a stack of matrices, the function must compute the number of non-zero singular values for each matrix in the stack.

        Parameters
        ----------
        x: _NAMESPACE_ARRAY
            input array having shape ``(..., M, N)`` and whose innermost two dimensions form ``MxN`` matrices. Should have a floating-point data type.
        rtol: Optional[Union[float, array]]
            relative tolerance for small singular values. Singular values approximately less than or equal to ``rtol * largest_singular_value`` are set to zero. If a ``float``, the value is equivalent to a zero-dimensional array having a real-valued floating-point data type determined by :ref:`type-promotion` (as applied to ``x``) and must be broadcast against each matrix. If an ``array``, must have a real-valued floating-point data type and must be compatible with ``shape(x)[:-2]`` (see :ref:`broadcasting`). If ``None``, the default value is ``max(M, N) * eps``, where ``eps`` must be the machine epsilon associated with the real-valued floating-point data type determined by :ref:`type-promotion` (as applied to ``x``). Default: ``None``.

        Returns
        -------
        out: _NAMESPACE_ARRAY
            an array containing the ranks. The returned array must have the default integer data type and must have shape ``(...)`` (i.e., must have a shape equal to ``shape(x)[:-2]``).

        Notes
        -----

        .. versionchanged:: 2022.12
        Added complex data type support.
        """

    @abstractmethod
    def matrix_transpose(
        self,
        x: _NAMESPACE_ARRAY, /
    ) -> _NAMESPACE_ARRAY:
        """Alias for :func:`~array_api.matrix_transpose`."""

    @abstractmethod
    def outer(
        self,
        x1: _NAMESPACE_ARRAY, x2: _NAMESPACE_ARRAY, /
    ) -> _NAMESPACE_ARRAY:
        """
        Returns the outer product of two vectors ``x1`` and ``x2``.

        Parameters
        ----------
        x1: _NAMESPACE_ARRAY
            first one-dimensional input array of size ``N``. Must have a numeric data type.
        x2: _NAMESPACE_ARRAY
            second one-dimensional input array of size ``M``. Must have a numeric data type.

        Returns
        -------
        out: _NAMESPACE_ARRAY
            a two-dimensional array containing the outer product and whose shape is ``(N, M)``. The returned array must have a data type determined by :ref:`type-promotion`.

        Notes
        -----

        .. versionchanged:: 2022.12
        Added complex data type support.
        """

    @abstractmethod
    def pinv(
        self,
        x: _NAMESPACE_ARRAY, /, *, rtol: Optional[Union[float, _NAMESPACE_ARRAY]] = None
    ) -> _NAMESPACE_ARRAY:
        r"""
        Returns the (Moore-Penrose) pseudo-inverse of a matrix (or a stack of matrices) ``x``.

        The pseudo-inverse of a matrix :math:`A`, denoted :math:`A^{+}`, is defined as the matrix that "solves" the least-squares problem :math:`Ax = b` (i.e., if :math:`\overline{x}` is a solution, then :math:`A^{+}` is the matrix such that :math:`\overline{x} = A^{+}b`).

        While the pseudo-inverse can be defined algebraically, one can understand the pseudo-inverse via singular value decomposition (SVD). Namely, if

        .. math::
        A = U \Sigma V^H

        is a singular decomposition of :math:`A`, then

        .. math::
        A^{+} = U \Sigma^{+} V^H

        where :math:`U` and :math:`V^H` are orthogonal matrices, :math:`\Sigma` is a diagonal matrix consisting of :math:`A`'s singular values, and :math:`\Sigma^{+}` is then a diagonal matrix consisting of the reciprocals of :math:`A`'s singular values, leaving zeros in place. During numerical computation, only elements larger than a small tolerance are considered nonzero, and all others replaced by zeros.

        When ``x`` is a stack of matrices, the function must compute the pseudo-inverse for each matrix in the stack.

        Parameters
        ----------
        x: _NAMESPACE_ARRAY
            input array having shape ``(..., M, N)`` and whose innermost two dimensions form ``MxN`` matrices. Should have a floating-point data type.
        rtol: Optional[Union[float, array]]
            relative tolerance for small singular values. Singular values approximately less than or equal to ``rtol * largest_singular_value`` are set to zero. If a ``float``, the value is equivalent to a zero-dimensional array having a real-valued floating-point data type determined by :ref:`type-promotion` (as applied to ``x``) and must be broadcast against each matrix. If an ``array``, must have a real-valued floating-point data type and must be compatible with ``shape(x)[:-2]`` (see :ref:`broadcasting`). If ``None``, the default value is ``max(M, N) * eps``, where ``eps`` must be the machine epsilon associated with the real-valued floating-point data type determined by :ref:`type-promotion` (as applied to ``x``). Default: ``None``.

        Returns
        -------
        out: _NAMESPACE_ARRAY
            an array containing the pseudo-inverse(s). The returned array must have a floating-point data type determined by :ref:`type-promotion` and must have shape ``(..., N, M)`` (i.e., must have the same shape as ``x``, except the innermost two dimensions must be transposed).

        Notes
        -----

        .. versionchanged:: 2022.12
        Added complex data type support.
        """

    @abstractmethod
    def qr(
        self,
        x: _NAMESPACE_ARRAY, /, *, mode: Literal["reduced", "complete"] = "reduced"
    ) -> Tuple[_NAMESPACE_ARRAY, _NAMESPACE_ARRAY]:
        r"""
        Returns the QR decomposition of a full column rank matrix (or a stack of matrices).

        If ``x`` is real-valued, let :math:`\mathbb{K}` be the set of real numbers :math:`\mathbb{R}`, and, if ``x`` is complex-valued, let :math:`\mathbb{K}` be the set of complex numbers :math:`\mathbb{C}`.

        The **complete QR decomposition** of a matrix :math:`x \in\ \mathbb{K}^{n \times n}` is defined as

        .. math::
        x = QR

        where :math:`Q \in\ \mathbb{K}^{m \times m}` is orthogonal when ``x`` is real-valued and unitary when ``x`` is complex-valued and where :math:`R \in\ \mathbb{K}^{m \times n}` is an upper triangular matrix with real diagonal (even when ``x`` is complex-valued).

        When :math:`m \gt n` (tall matrix), as :math:`R` is upper triangular, the last :math:`m - n` rows are zero. In this case, the last :math:`m - n` columns of :math:`Q` can be dropped to form the **reduced QR decomposition**.

        .. math::
        x = QR

        where :math:`Q \in\ \mathbb{K}^{m \times n}` and :math:`R \in\ \mathbb{K}^{n \times n}`.

        The reduced QR decomposition equals with the complete QR decomposition when :math:`n \geq m` (wide matrix).

        When ``x`` is a stack of matrices, the function must compute the QR decomposition for each matrix in the stack.

        .. note::
        Whether an array library explicitly checks whether an input array is a full column rank matrix (or a stack of full column rank matrices) is implementation-defined.

        .. warning::
        The elements in the diagonal of :math:`R` are not necessarily positive. Accordingly, the returned QR decomposition is only unique up to the sign of the diagonal of :math:`R`, and different libraries or inputs on different devices may produce different valid decompositions.

        .. warning::
        The QR decomposition is only well-defined if the first ``k = min(m,n)`` columns of every matrix in ``x`` are linearly independent.

        Parameters
        ----------
        x: _NAMESPACE_ARRAY
            input array having shape ``(..., M, N)`` and whose innermost two dimensions form ``MxN`` matrices of rank ``N``. Should have a floating-point data type.
        mode: Literal['reduced', 'complete']
            decomposition mode. Should be one of the following modes:

            -   ``'reduced'``: compute only the leading ``K`` columns of ``q``, such that ``q`` and ``r`` have dimensions ``(..., M, K)`` and ``(..., K, N)``, respectively, and where ``K = min(M, N)``.
            -   ``'complete'``: compute ``q`` and ``r`` with dimensions ``(..., M, M)`` and ``(..., M, N)``, respectively.

            Default: ``'reduced'``.

        Returns
        -------
        out: Tuple[array, array]
            a namedtuple ``(Q, R)`` whose

            -   first element must have the field name ``Q`` and must be an array whose shape depends on the value of ``mode`` and contain matrices with orthonormal columns. If ``mode`` is ``'complete'``, the array must have shape ``(..., M, M)``. If ``mode`` is ``'reduced'``, the array must have shape ``(..., M, K)``, where ``K = min(M, N)``. The first ``x.ndim-2`` dimensions must have the same size as those of the input array ``x``.
            -   second element must have the field name ``R`` and must be an array whose shape depends on the value of ``mode`` and contain upper-triangular matrices. If ``mode`` is ``'complete'``, the array must have shape ``(..., M, N)``. If ``mode`` is ``'reduced'``, the array must have shape ``(..., K, N)``, where ``K = min(M, N)``. The first ``x.ndim-2`` dimensions must have the same size as those of the input ``x``.

            Each returned array must have a floating-point data type determined by :ref:`type-promotion`.

        Notes
        -----

        .. versionchanged:: 2022.12
        Added complex data type support.
        """

    @abstractmethod
    def slogdet(
        self,
        x: _NAMESPACE_ARRAY, /
    ) -> Tuple[_NAMESPACE_ARRAY, _NAMESPACE_ARRAY]:
        r"""
        Returns the sign and the natural logarithm of the absolute value of the determinant of a square matrix (or a stack of square matrices) ``x``.

        .. note::
        The purpose of this function is to calculate the determinant more accurately when the determinant is either very small or very large, as calling ``det`` may overflow or underflow.

        The sign of the determinant is given by

        .. math::
        \operatorname{sign}(\det x) = \begin{cases}
        0 & \textrm{if } \det x = 0 \\
        \frac{\det x}{|\det x|} & \textrm{otherwise}
        \end{cases}

        where :math:`|\det x|` is the absolute value of the determinant of ``x``.

        When ``x`` is a stack of matrices, the function must compute the sign and natural logarithm of the absolute value of the determinant for each matrix in the stack.

        **Special Cases**

        For real-valued floating-point operands,

        - If the determinant is zero, the ``sign`` should be ``0`` and ``logabsdet`` should be ``-infinity``.

        For complex floating-point operands,

        - If the determinant is ``0 + 0j``, the ``sign`` should be ``0 + 0j`` and ``logabsdet`` should be ``-infinity + 0j``.

        .. note::
        Depending on the underlying algorithm, when the determinant is zero, the returned result may differ from ``-infinity`` (or ``-infinity + 0j``). In all cases, the determinant should be equal to ``sign * exp(logabsdet)`` (although, again, the result may be subject to numerical precision errors).

        Parameters
        ----------
        x: _NAMESPACE_ARRAY
            input array having shape ``(..., M, M)`` and whose innermost two dimensions form square matrices. Should have a floating-point data type.

        Returns
        -------
        out: Tuple[array, array]
            a namedtuple (``sign``, ``logabsdet``) whose

            -   first element must have the field name ``sign`` and must be an array containing a number representing the sign of the determinant for each square matrix. Must have the same data type as ``x``.
            -   second element must have the field name ``logabsdet`` and must be an array containing the natural logarithm of the absolute value of the determinant for each square matrix. If ``x`` is real-valued, the returned array must have a real-valued floating-point data type determined by :ref:`type-promotion`. If ``x`` is complex, the returned array must have a real-valued floating-point data type having the same precision as ``x`` (e.g., if ``x`` is ``complex64``, ``logabsdet`` must have a ``float32`` data type).

            Each returned array must have shape ``shape(x)[:-2]``.

        Notes
        -----

        .. versionchanged:: 2022.12
        Added complex data type support.
        """

    @abstractmethod
    def solve(
        self,
        x1: _NAMESPACE_ARRAY, x2: _NAMESPACE_ARRAY, /
    ) -> _NAMESPACE_ARRAY:
        r"""
        Returns the solution of a square system of linear equations with a unique solution.

        Let ``x1`` equal :math:`A` and ``x2`` equal :math:`B`. If the promoted data type of ``x1`` and ``x2`` is real-valued, let :math:`\mathbb{K}` be the set of real numbers :math:`\mathbb{R}`, and, if the promoted data type of ``x1`` and ``x2`` is complex-valued, let :math:`\mathbb{K}` be the set of complex numbers :math:`\mathbb{C}`.

        This function computes the solution :math:`X \in\ \mathbb{K}^{m \times k}` of the **linear system** associated to :math:`A \in\ \mathbb{K}^{m \times m}` and :math:`B \in\ \mathbb{K}^{m \times k}` and is defined as

        .. math::
        AX = B

        This system of linear equations has a unique solution if and only if :math:`A` is invertible.

        .. note::
        Whether an array library explicitly checks whether ``x1`` is invertible is implementation-defined.

        When ``x1`` and/or ``x2`` is a stack of matrices, the function must compute a solution for each matrix in the stack.

        Parameters
        ----------
        x1: _NAMESPACE_ARRAY
            coefficient array ``A`` having shape ``(..., M, M)`` and whose innermost two dimensions form square matrices. Must be of full rank (i.e., all rows or, equivalently, columns must be linearly independent). Should have a floating-point data type.
        x2: _NAMESPACE_ARRAY
            ordinate (or "dependent variable") array ``B``. If ``x2`` has shape ``(M,)``, ``x2`` is equivalent to an array having shape ``(..., M, 1)``. If ``x2`` has shape ``(..., M, K)``, each column ``k`` defines a set of ordinate values for which to compute a solution, and ``shape(x2)[:-2]`` must be compatible with ``shape(x1)[:-2]`` (see :ref:`broadcasting`). Should have a floating-point data type.

        Returns
        -------
        out: _NAMESPACE_ARRAY
            an array containing the solution to the system ``AX = B`` for each square matrix. If ``x2`` has shape ``(M,)``, the returned array must have shape equal to ``shape(x1)[:-2] + shape(x2)[-1:]``. Otherwise, if ``x2`` has shape ``(..., M, K)```, the returned array must have shape equal to ``(..., M, K)``, where ``...`` refers to the result of broadcasting ``shape(x1)[:-2]`` and ``shape(x2)[:-2]``. The returned array must have a floating-point data type determined by :ref:`type-promotion`.

        Notes
        -----

        .. versionchanged:: 2022.12
        Added complex data type support.

        .. versionchanged:: 2024.12
        Clarified broadcasting semantics and the shape of the output array.
        """

    @abstractmethod
    def svd(
        self,
        x: _NAMESPACE_ARRAY, /, *, full_matrices: bool = True
    ) -> Tuple[_NAMESPACE_ARRAY, _NAMESPACE_ARRAY, _NAMESPACE_ARRAY]:
        r"""
        Returns a singular value decomposition (SVD) of a matrix (or a stack of matrices) ``x``.

        If ``x`` is real-valued, let :math:`\mathbb{K}` be the set of real numbers :math:`\mathbb{R}`, and, if ``x`` is complex-valued, let :math:`\mathbb{K}` be the set of complex numbers :math:`\mathbb{C}`.

        The full **singular value decomposition** of an :math:`m \times n` matrix :math:`x \in\ \mathbb{K}^{m \times n}` is a factorization of the form

        .. math::
        x = U \Sigma V^H

        where :math:`U \in\ \mathbb{K}^{m \times m}`, :math:`\Sigma \in\ \mathbb{K}^{m \times\ n}`, :math:`\operatorname{diag}(\Sigma) \in\ \mathbb{R}^{k}` with :math:`k = \operatorname{min}(m, n)`, :math:`V^H \in\ \mathbb{K}^{n \times n}`, and where :math:`V^H` is the conjugate transpose when :math:`V` is complex and the transpose when :math:`V` is real-valued. When ``x`` is real-valued, :math:`U`, :math:`V` (and thus :math:`V^H`) are orthogonal, and, when ``x`` is complex, :math:`U`, :math:`V` (and thus :math:`V^H`) are unitary.

        When :math:`m \gt n` (tall matrix), we can drop the last :math:`m - n` columns of :math:`U` to form the reduced SVD

        .. math::
        x = U \Sigma V^H

        where :math:`U \in\ \mathbb{K}^{m \times k}`, :math:`\Sigma \in\ \mathbb{K}^{k \times\ k}`, :math:`\operatorname{diag}(\Sigma) \in\ \mathbb{R}^{k}`, and :math:`V^H \in\ \mathbb{K}^{k \times n}`. In this case, :math:`U` and :math:`V` have orthonormal columns.

        Similarly, when :math:`n \gt m` (wide matrix), we can drop the last :math:`n - m` columns of :math:`V` to also form a reduced SVD.

        This function returns the decomposition :math:`U`, :math:`S`, and :math:`V^H`, where :math:`S = \operatorname{diag}(\Sigma)`.

        When ``x`` is a stack of matrices, the function must compute the singular value decomposition for each matrix in the stack.

        .. warning::
        The returned arrays :math:`U` and :math:`V` are neither unique nor continuous with respect to ``x``. Because :math:`U` and :math:`V` are not unique, different hardware and software may compute different singular vectors.

        Non-uniqueness stems from the fact that multiplying any pair of singular vectors :math:`u_k`, :math:`v_k` by :math:`-1` when ``x`` is real-valued and by :math:`e^{\phi j}` (:math:`\phi \in \mathbb{R}`) when ``x`` is complex produces another two valid singular vectors of the matrix.

        Parameters
        ----------
        x: _NAMESPACE_ARRAY
            input array having shape ``(..., M, N)`` and whose innermost two dimensions form matrices on which to perform singular value decomposition. Should have a floating-point data type.
        full_matrices: bool
            If ``True``, compute full-sized ``U`` and ``Vh``, such that ``U`` has shape ``(..., M, M)`` and ``Vh`` has shape ``(..., N, N)``. If ``False``, compute on the leading ``K`` singular vectors, such that ``U`` has shape ``(..., M, K)`` and ``Vh`` has shape ``(..., K, N)`` and where ``K = min(M, N)``. Default: ``True``.

        Returns
        -------
        out: Tuple[array, array, array]
            a namedtuple ``(U, S, Vh)`` whose

            -   first element must have the field name ``U`` and must be an array whose shape depends on the value of ``full_matrices`` and contain matrices with orthonormal columns (i.e., the columns are left singular vectors). If ``full_matrices`` is ``True``, the array must have shape ``(..., M, M)``. If ``full_matrices`` is ``False``, the array must have shape ``(..., M, K)``, where ``K = min(M, N)``. The first ``x.ndim-2`` dimensions must have the same shape as those of the input ``x``. Must have the same data type as ``x``.
            -   second element must have the field name ``S`` and must be an array with shape ``(..., K)`` that contains the vector(s) of singular values of length ``K``, where ``K = min(M, N)``. For each vector, the singular values must be sorted in descending order by magnitude, such that ``s[..., 0]`` is the largest value, ``s[..., 1]`` is the second largest value, et cetera. The first ``x.ndim-2`` dimensions must have the same shape as those of the input ``x``. Must have a real-valued floating-point data type having the same precision as ``x`` (e.g., if ``x`` is ``complex64``, ``S`` must have a ``float32`` data type).
            -   third element must have the field name ``Vh`` and must be an array whose shape depends on the value of ``full_matrices`` and contain orthonormal rows (i.e., the rows are the right singular vectors and the array is the adjoint). If ``full_matrices`` is ``True``, the array must have shape ``(..., N, N)``. If ``full_matrices`` is ``False``, the array must have shape ``(..., K, N)`` where ``K = min(M, N)``. The first ``x.ndim-2`` dimensions must have the same shape as those of the input ``x``. Must have the same data type as ``x``.

        Notes
        -----

        .. versionchanged:: 2022.12
        Added complex data type support.
        """

    @abstractmethod
    def svdvals(
        self,
        x: _NAMESPACE_ARRAY, /
    ) -> _NAMESPACE_ARRAY:
        """
        Returns the singular values of a matrix (or a stack of matrices) ``x``.

        When ``x`` is a stack of matrices, the function must compute the singular values for each matrix in the stack.

        Parameters
        ----------
        x: _NAMESPACE_ARRAY
            input array having shape ``(..., M, N)`` and whose innermost two dimensions form matrices on which to perform singular value decomposition. Should have a floating-point data type.

        Returns
        -------
        out: _NAMESPACE_ARRAY
            an array with shape ``(..., K)`` that contains the vector(s) of singular values of length ``K``, where ``K = min(M, N)``. For each vector, the singular values must be sorted in descending order by magnitude, such that ``s[..., 0]`` is the largest value, ``s[..., 1]`` is the second largest value, et cetera. The first ``x.ndim-2`` dimensions must have the same shape as those of the input ``x``. The returned array must have a real-valued floating-point data type having the same precision as ``x`` (e.g., if ``x`` is ``complex64``, the returned array must have a ``float32`` data type).

        Notes
        -----

        .. versionchanged:: 2022.12
        Added complex data type support.
        """

    @abstractmethod
    def tensordot(
        self,
        x1: _NAMESPACE_ARRAY,
        x2: _NAMESPACE_ARRAY,
        /,
        *,
        axes: Union[int, Tuple[Sequence[int], Sequence[int]]] = 2,
    ) -> _NAMESPACE_ARRAY:
        """Alias for :func:`~array_api.tensordot`."""

    @abstractmethod
    def trace(
        self,
        x: _NAMESPACE_ARRAY, /, *, offset: int = 0, dtype: Optional[_NAMESPACE_DTYPE] = None
    ) -> _NAMESPACE_ARRAY:
        """
        Returns the sum along the specified diagonals of a matrix (or a stack of matrices) ``x``.

        Parameters
        ----------
        x: _NAMESPACE_ARRAY
            input array having shape ``(..., M, N)`` and whose innermost two dimensions form ``MxN`` matrices. Should have a numeric data type.
        offset: int
            offset specifying the off-diagonal relative to the main diagonal.

            -   ``offset = 0``: the main diagonal.
            -   ``offset > 0``: off-diagonal above the main diagonal.
            -   ``offset < 0``: off-diagonal below the main diagonal.

            Default: ``0``.
        dtype: Optional[dtype]
            data type of the returned array. If ``None``, the returned array must have the same data type as ``x``, unless ``x`` has an integer data type supporting a smaller range of values than the default integer data type (e.g., ``x`` has an ``int16`` or ``uint32`` data type and the default integer data type is ``int64``). In those latter cases:

            -   if ``x`` has a signed integer data type (e.g., ``int16``), the returned array must have the default integer data type.
            -   if ``x`` has an unsigned integer data type (e.g., ``uint16``), the returned array must have an unsigned integer data type having the same number of bits as the default integer data type (e.g., if the default integer data type is ``int32``, the returned array must have a ``uint32`` data type).

            If the data type (either specified or resolved) differs from the data type of ``x``, the input array should be cast to the specified data type before computing the sum (rationale: the ``dtype`` keyword argument is intended to help prevent overflows). Default: ``None``.

        Returns
        -------
        out: _NAMESPACE_ARRAY
            an array containing the traces and whose shape is determined by removing the last two dimensions and storing the traces in the last array dimension. For example, if ``x`` has rank ``k`` and shape ``(I, J, K, ..., L, M, N)``, then an output array has rank ``k-2`` and shape ``(I, J, K, ..., L)`` where

            ::

            out[i, j, k, ..., l] = trace(a[i, j, k, ..., l, :, :])

            The returned array must have a data type as described by the ``dtype`` parameter above.

        Notes
        -----

        **Special Cases**

        Let ``N`` equal the number of elements over which to compute the sum.

        -   If ``N`` is ``0``, the sum is ``0`` (i.e., the empty sum).

        For both real-valued and complex floating-point operands, special cases must be handled as if the operation is implemented by successive application of :func:`~array_api.add`.

        .. versionchanged:: 2022.12
        Added complex data type support.

        .. versionchanged:: 2023.12
        Required the function to return a floating-point array having the same data type as the input array when provided a floating-point array.
        """

    @abstractmethod
    def vecdot(
        self,
        x1: _NAMESPACE_ARRAY, x2: _NAMESPACE_ARRAY, /, *, axis: int = -1
    ) -> _NAMESPACE_ARRAY:
        """Alias for :func:`~array_api.vecdot`."""

    @abstractmethod
    def vector_norm(
        self,
        x: _NAMESPACE_ARRAY,
        /,
        *,
        axis: Optional[Union[int, Tuple[int, ...]]] = None,
        keepdims: bool = False,
        ord: Union[int, float] = 2,
    ) -> _NAMESPACE_ARRAY:
        r"""
        Computes the vector norm of a vector (or batch of vectors) ``x``.

        Parameters
        ----------
        x: _NAMESPACE_ARRAY
            input array. Should have a floating-point data type.
        axis: Optional[Union[int, Tuple[int, ...]]]
            If an integer, ``axis`` specifies the axis (dimension) along which to compute vector norms. If an n-tuple, ``axis`` specifies the axes (dimensions) along which to compute batched vector norms. If ``None``, the vector norm must be computed over all array values (i.e., equivalent to computing the vector norm of a flattened array). Negative indices must be supported. Default: ``None``.
        keepdims: bool
            If ``True``, the axes (dimensions) specified by ``axis`` must be included in the result as singleton dimensions, and, accordingly, the result must be compatible with the input array (see :ref:`broadcasting`). Otherwise, if ``False``, the axes (dimensions) specified by ``axis`` must not be included in the result. Default: ``False``.
        ord: Union[int, float, Literal[inf, -inf]]
            order of the norm. The following mathematical norms must be supported:

            +------------------+----------------------------+
            | ord              | description                |
            +==================+============================+
            | 1                | L1-norm (Manhattan)        |
            +------------------+----------------------------+
            | 2                | L2-norm (Euclidean)        |
            +------------------+----------------------------+
            | inf              | infinity norm              |
            +------------------+----------------------------+
            | (int,float >= 1) | p-norm                     |
            +------------------+----------------------------+

            The following non-mathematical "norms" must be supported:

            +------------------+--------------------------------+
            | ord              | description                    |
            +==================+================================+
            | 0                | sum(a != 0)                    |
            +------------------+--------------------------------+
            | -1               | 1./sum(1./abs(a))              |
            +------------------+--------------------------------+
            | -2               | 1./sqrt(sum(1./abs(a)\*\*2))   |
            +------------------+--------------------------------+
            | -inf             | min(abs(a))                    |
            +------------------+--------------------------------+
            | (int,float < 1)  | sum(abs(a)\*\*ord)\*\*(1./ord) |
            +------------------+--------------------------------+

            Default: ``2``.

        Returns
        -------
        out: _NAMESPACE_ARRAY
            an array containing the vector norms. If ``axis`` is ``None``, the returned array must be a zero-dimensional array containing a vector norm. If ``axis`` is a scalar value (``int`` or ``float``), the returned array must have a rank which is one less than the rank of ``x``. If ``axis`` is a ``n``-tuple, the returned array must have a rank which is ``n`` less than the rank of ``x``. If ``x`` has a real-valued data type, the returned array must have a real-valued floating-point data type determined by :ref:`type-promotion`. If ``x`` has a complex-valued data type, the returned array must have a real-valued floating-point data type whose precision matches the precision of ``x`` (e.g., if ``x`` is ``complex128``, then the returned array must have a ``float64`` data type).

        Notes
        -----

        .. versionchanged:: 2022.12
        Added complex data type support.
        """


from array_api_typing.typing_compat import ArrayAPIArray, ArrayAPIDevice, ArrayAPIDType, SupportsDLPack, SetIndex, GetIndex
from ._api_typing import ArrayAPINamespace

__all__ = [
    "ArrayAPIArray",
    "ArrayAPIDevice",
    "ArrayAPIDType",
    "SetIndex",
    "GetIndex",
    "SupportsDLPack",
    "ArrayAPINamespace"
]


from typing import Union, Optional, Tuple, Any
from ._typing import ARRAY_TYPE, DTYPE_TYPE, DEVICE_TYPE, RNG_TYPE
import numpy as np
import jax

__all__ = [
    "random_number_generator",
    "random_discrete_uniform",
    "random_uniform",
    "random_exponential",
    "random_normal",
    "random_geometric",
    "random_permutation"
]

def random_number_generator(
    seed : Optional[int] = None,
    *,
    device : Optional[DEVICE_TYPE] = None
) -> RNG_TYPE:
    rng_seed = np.random.randint(65535) if seed is None else seed
    rng = jax.random.key(
        seed=rng_seed
    )
    return rng

def random_discrete_uniform(
    shape : Union[int, Tuple[int, ...]], 
    /,
    from_num : int, 
    to_num : int, 
    *,
    rng : RNG_TYPE, 
    dtype : Optional[DTYPE_TYPE] = None, 
    device : Optional[DEVICE_TYPE] = None
) -> Tuple[RNG_TYPE, ARRAY_TYPE]:
    new_rng, rng = jax.random.split(rng)
    t = jax.random.randint(rng, shape, minval=int(from_num), maxval=int(to_num), dtype=dtype or int)
    if device is not None:
        t = jax.device_put(t, device)
    return new_rng, t

def random_uniform(
    shape: Union[int, Tuple[int, ...]], 
    /,
    *,
    rng : RNG_TYPE, 
    low : float = 0.0, high : float = 1.0,
    dtype : Optional[DTYPE_TYPE] = None, 
    device : Optional[DEVICE_TYPE] = None
) -> Tuple[RNG_TYPE, ARRAY_TYPE]:
    new_rng, rng = jax.random.split(rng)
    data = jax.random.uniform(rng, shape, dtype=dtype or float, minval=low, maxval=high)
    if device is not None:
        data = jax.device_put(data, device)
    return new_rng, data

def random_exponential(
    shape: Union[int, Tuple[int, ...]], 
    /,
    *,
    rng : RNG_TYPE,  
    lambd : float = 1.0, 
    dtype : Optional[DTYPE_TYPE] = None, 
    device : Optional[DEVICE_TYPE] = None
) -> Tuple[RNG_TYPE, ARRAY_TYPE]:
    new_rng, rng = jax.random.split(rng)
    data = jax.random.exponential(rng, shape, dtype=dtype or float) / lambd
    if device is not None:
        data = jax.device_put(data, device)
    return new_rng, data

def random_normal(
    shape: Union[int, Tuple[int, ...]],
    /,
    *, 
    rng : RNG_TYPE,
    mean : float = 0.0, std : float = 1.0, 
    dtype : Optional[DTYPE_TYPE] = None, 
    device : Optional[Any] = None
) -> Tuple[RNG_TYPE, ARRAY_TYPE]:
    new_rng, rng = jax.random.split(rng)
    data = jax.random.normal(rng, shape, dtype=dtype or float) * std + mean
    if device is not None:
        data = jax.device_put(data, device)
    return new_rng, data

def random_geometric(
    shape: Union[int, Tuple[int, ...]], 
    /,
    *,
    p: float, 
    rng: RNG_TYPE,
    dtype: Optional[DTYPE_TYPE] = None, 
    device: Optional[Any] = None
) -> Tuple[RNG_TYPE, ARRAY_TYPE]:
    new_rng, rng = jax.random.split(rng)
    data = jax.random.geometric(rng, p=p, shape=shape, dtype=dtype or int)
    if device is not None:
        data = jax.device_put(data, device)
    return new_rng, data

def random_permutation(
    n : int,
    /,
    *,
    rng: RNG_TYPE,
    dtype: Optional[DTYPE_TYPE] = None,
    device: Optional[DEVICE_TYPE] = None
) -> Tuple[RNG_TYPE, ARRAY_TYPE]:
    new_rng, rng = jax.random.split(rng)
    data = jax.random.permutation(rng, n, dtype=dtype or int)
    if device is not None:
        data = jax.device_put(data, device)
    return new_rng, data


from array_api_compat.torch import *
from array_api_compat.torch import __array_api_version__, __array_namespace_info__
from array_api_compat.common._helpers import *

simplified_name = "pytorch"

from array_api_compat import torch as compat_module

# Import and bind all functions from array_api_extra before exposing them
import array_api_extra
from functools import partial
for api_name in dir(array_api_extra):
    if api_name.startswith('_'):
        continue

    if api_name in ['at', 'broadcast_shapes']:
        globals()[api_name] = getattr(array_api_extra, api_name)
    else:
        globals()[api_name] = partial(
            getattr(array_api_extra, api_name),
            xp=compat_module
        )

from ._typing import *
from ._extra import *
__import__(__package__ + ".random")


from typing import Protocol, TypeVar, Optional, Any, Tuple, Union, Type, TypedDict, List, TypeAlias
from dataclasses import dataclass
from abc import abstractmethod
from ._array_typing import Device, DType

__all__ = [
    "SupportsBufferProtocol",
    "DefaultDataTypes",
    "DataTypes",
    "Capabilities",
    "finfo_object",
    "iinfo_object",
    "NestedSequence",
    "Info"
]

SupportsBufferProtocol = TypeVar("SupportsBufferProtocol")

"""
https://github.com/data-apis/array-api/blob/main/src/array_api_stubs/_2024_12/_types.py#L55
"""
DefaultDataTypes = TypedDict(
    "DefaultDataTypes",
    {
        "real floating": DType,
        "complex floating": DType,
        "integral": DType,
        "indexing": DType,
    },
)
DataTypes = TypedDict(
    "DataTypes",
    {
        "bool": DType,
        "float32": DType,
        "float64": DType,
        "complex64": DType,
        "complex128": DType,
        "int8": DType,
        "int16": DType,
        "int32": DType,
        "int64": DType,
        "uint8": DType,
        "uint16": DType,
        "uint32": DType,
        "uint64": DType,
    },
    total=False,
)
Capabilities = TypedDict(
    "Capabilities",
    {
        "boolean indexing": bool,
        "data-dependent shapes": bool,
        "max rank": Optional[int],
    },
)

@dataclass
class finfo_object:
    """Dataclass returned by `finfo`."""
    bits: int
    eps: float
    max: float
    min: float
    smallest_normal: float
    dtype: DType

@dataclass
class iinfo_object:
    """Dataclass returned by `iinfo`."""
    bits: int
    max: int
    min: int
    dtype: DType

_T_co = TypeVar("_T_co", covariant=True)
class NestedSequence(Protocol[_T_co]):
    def __getitem__(self, key: int, /) -> Union[_T_co, 'NestedSequence[_T_co]']:
        ...

    def __len__(self, /) -> int:
        ...

class Info(Protocol):
    """Namespace returned by `__array_namespace_info__`."""

    def capabilities(self) -> Capabilities:
        ...

    def default_device(self) -> Device:
        ...

    def default_dtypes(self, *, device: Optional[Device]) -> DefaultDataTypes:
        ...

    def devices(self) -> List[Device]:
        ...

    def dtypes(
        self, *, device: Optional[Device] = None, kind: Optional[Union[str, Tuple[str, ...]]] = None
    ) -> DataTypes: # In the original signature there's no default parameters for `device` and `kind`, but we add them since they exist in `array_api_compat` modules
        ...



from ._cls_base import ComputeBackendImplCls
from ._implementations import jax as jax_impl

class JaxComputeBackend(metaclass=ComputeBackendImplCls[jax_impl.ARRAY_TYPE, jax_impl.DEVICE_TYPE, jax_impl.DTYPE_TYPE, jax_impl.RNG_TYPE]):
    ARRAY_TYPE = jax_impl.ARRAY_TYPE
    DTYPE_TYPE = jax_impl.DTYPE_TYPE
    DEVICE_TYPE = jax_impl.DEVICE_TYPE
    RNG_TYPE = jax_impl.RNG_TYPE

for name in dir(jax_impl):
    if not name.startswith('_') or name in [
        '__array_namespace_info__',
        '__array_api_version__',
    ]:
        setattr(JaxComputeBackend, name, getattr(jax_impl, name))

__all__ = [
    'JaxComputeBackend',
]


from typing import Any, Union, Optional
import numpy as np
from ._typing import ARRAY_TYPE, DTYPE_TYPE, DEVICE_TYPE, RNG_TYPE
from xbarray.backends.base import ComputeBackend, SupportsDLPack

__all__ = [
    "default_integer_dtype",
    "default_floating_dtype",
    "default_boolean_dtype",
    "is_backendarray",
    "from_numpy",
    "from_other_backend",
    "to_numpy",
    "to_dlpack",
    "dtype_is_real_integer",
    "dtype_is_real_floating",
    "dtype_is_boolean",
    "abbreviate_array",
    "map_fn_over_arrays",
]

default_integer_dtype = int
default_floating_dtype = float
default_boolean_dtype = bool

def is_backendarray(data : Any) -> bool:
    return isinstance(data, np.ndarray)

def from_numpy(
    data : np.ndarray,
    /,
    *,
    dtype : Optional[DTYPE_TYPE] = None,
    device : Optional[DEVICE_TYPE] = None
) -> ARRAY_TYPE:
    return data

def from_other_backend(
    other_backend: ComputeBackend,
    data: Any,
    /,
) -> ARRAY_TYPE:
    return other_backend.to_numpy(data)

def to_numpy(
    data : ARRAY_TYPE
) -> np.ndarray:
    return data

def to_dlpack(
    data: ARRAY_TYPE,
    /,
) -> SupportsDLPack:
    return data

def dtype_is_real_integer(
    dtype: DTYPE_TYPE
) -> bool:
    return np.issubdtype(dtype, np.integer)

def dtype_is_real_floating(
    dtype: DTYPE_TYPE
) -> bool:
    return np.issubdtype(dtype, np.floating)

def dtype_is_boolean(
    dtype: DTYPE_TYPE
) -> bool:
    return dtype == np.bool_ or dtype == bool

from .._common.implementations import *
from array_api_compat import numpy as compat_module
abbreviate_array = get_abbreviate_array_function(
    backend=compat_module,
    default_integer_dtype=default_integer_dtype,
    func_dtype_is_real_floating=dtype_is_real_floating,
    func_dtype_is_real_integer=dtype_is_real_integer,
    func_dtype_is_boolean=dtype_is_boolean,
)

map_fn_over_arrays = get_map_fn_over_arrays_function(
    is_backendarray=is_backendarray,
)

pad_dim = get_pad_dim_function(
    backend=compat_module,
)


from typing import Optional, Generic, TypeVar, Dict, Union, Any, Sequence, SupportsFloat, Tuple, Type, Callable, Mapping, Protocol
import abc
from array_api_typing.typing_extra import *
import numpy as np

ArrayAPISetIndex = SetIndex
ArrayAPIGetIndex = GetIndex

__all__ = [
    "RNGBackend",
    "ComputeBackend",
    "BArrayType",
    "BDeviceType",
    "BDtypeType",
    "BRNGType",
    "SupportsDLPack",
    "ArrayAPIArray",
    "ArrayAPIDevice",
    "ArrayAPIDType",
    "ArrayAPINamespace",
    "ArrayAPISetIndex",
    "ArrayAPIGetIndex",
]

BArrayType = TypeVar("BArrayType", covariant=True, bound=ArrayAPIArray)
BDeviceType = TypeVar("BDeviceType", covariant=True, bound=ArrayAPIDevice)
BDtypeType = TypeVar("BDtypeType", covariant=True, bound=ArrayAPIDType)
BRNGType = TypeVar("BRNGType", covariant=True)
class RNGBackend(Protocol[BArrayType, BDeviceType, BDtypeType, BRNGType]):
    @abc.abstractmethod
    def random_number_generator(
        self, 
        seed : Optional[int] = None, 
        *,
        device : Optional[BDeviceType] = None
    ) -> BRNGType:
        raise NotImplementedError
    
    @abc.abstractmethod
    def random_discrete_uniform(
        self, 
        shape : Union[int, Tuple[int, ...]], 
        /,
        from_num : int, 
        to_num : int,
        *,
        rng : BRNGType, 
        dtype : Optional[BDtypeType] = None, 
        device : Optional[BDeviceType] = None, 
    ) -> Tuple[BRNGType, BArrayType]:
        """
        Sample from a discrete uniform distribution [from_num, to_num) with shape `shape`.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def random_uniform(
        self, 
        shape: Union[int, Tuple[int, ...]], 
        /,
        *,
        rng : BRNGType, 
        low : float = 0.0, high : float = 1.0, 
        dtype : Optional[BDtypeType] = None, 
        device : Optional[BDeviceType] = None,
    ) -> Tuple[BRNGType, BArrayType]:
        raise NotImplementedError

    @abc.abstractmethod
    def random_exponential(
        self, 
        shape: Union[int, Tuple[int, ...]], 
        /,
        *,
        rng : BRNGType, 
        lambd : float = 1.0, 
        dtype : Optional[BDtypeType] = None, 
        device : Optional[BDeviceType] = None, 
    ) -> Tuple[BRNGType, BArrayType]:
        raise NotImplementedError

    @abc.abstractmethod
    def random_normal(
        self, 
        shape: Union[int, Tuple[int, ...]], 
        /,
        *,
        rng : BRNGType, 
        mean : float = 0.0, std : float = 1.0, 
        dtype : Optional[BDtypeType] = None, 
        device : Optional[BDeviceType] = None,
    ) -> Tuple[BRNGType, BArrayType]:
        raise NotImplementedError

    @abc.abstractmethod
    def random_geometric(
        self, 
        shape: Union[int, Tuple[int, ...]], 
        /,
        *, 
        p : float, 
        rng : BRNGType, 
        dtype : Optional[BDtypeType] = None, 
        device : Optional[BDeviceType] = None,
    ) -> Tuple[BRNGType, BArrayType]:
        raise NotImplementedError

    @abc.abstractmethod
    def random_permutation(
        self, 
        n : int, 
        /,
        *, 
        rng : BRNGType, 
        dtype : Optional[BDtypeType] = None, 
        device : Optional[BDeviceType] = None
    ) -> Tuple[BRNGType, BArrayType]:
        raise NotImplementedError

class ComputeBackend(ArrayAPINamespace[BArrayType, BDeviceType, BDtypeType], Protocol[BArrayType, BDeviceType, BDtypeType, BRNGType]):
    simplified_name : str
    ARRAY_TYPE : Type[BArrayType]
    DEVICE_TYPE : Type[BDeviceType]
    DTYPE_TYPE : Type[BDtypeType]
    RNG_TYPE : Type[BRNGType]
    default_integer_dtype : BDtypeType
    default_floating_dtype : BDtypeType
    default_boolean_dtype : BDtypeType
    random : RNGBackend[BArrayType, BDeviceType, BDtypeType, BRNGType]

    @abc.abstractmethod
    def is_backendarray(self, data : Any) -> bool:
        raise NotImplementedError
    
    @abc.abstractmethod
    def from_numpy(
        self, 
        data : np.ndarray, 
        /,
        *,
        dtype : Optional[BDtypeType] = None, 
        device : Optional[BDeviceType] = None
    ) -> BArrayType:
        raise NotImplementedError
    
    @abc.abstractmethod
    def from_other_backend(
        self,
        other_backend : "ComputeBackend",
        data : ArrayAPIArray,
        /
    ) -> BArrayType:
        """
        Convert an array from another backend to this backend.
        The other backend must be compatible with the ArrayAPI.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def to_numpy(
        self, 
        data : BArrayType
    ) -> np.ndarray:
        raise NotImplementedError

    @abc.abstractmethod
    def to_dlpack(
        self, 
        data : BArrayType, 
        /,
    ) -> SupportsDLPack:
        raise NotImplementedError

    @abc.abstractmethod
    def dtype_is_real_integer(self, dtype : BDtypeType) -> bool:
        raise NotImplementedError
    
    @abc.abstractmethod
    def dtype_is_real_floating(self, dtype : BDtypeType) -> bool:
        raise NotImplementedError
    
    @abc.abstractmethod
    def dtype_is_boolean(self, dtype : BDtypeType) -> bool:
        raise NotImplementedError
    
    @abc.abstractmethod
    def abbreviate_array(self, x : BArrayType, try_cast_scalar: bool = True) -> Union[float, int, BArrayType]:
        """
        Abbreviates an array to a single element if possible.
        Or, if some dimensions are the same, abbreviates to a smaller array (but with the same number of dimensions).
        """
        raise NotImplementedError
    
    @abc.abstractmethod
    def map_fn_over_arrays(self, data : Any, func : Callable[[BArrayType], BArrayType]) -> Any:
        """
        Map a function over arrays in a data structure and produce a new data structure with the same shape.
        This is useful for applying a function to all arrays in a nested structure.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def pad_dim(self, x : BArrayType, dim : int, target_size : int, value : Union[float, int] = 0) -> BArrayType:
        """
        Pad a dimension of an array to a target size with a given value.
        If the dimension is already the target size, return the original array.
        If the dimension is larger than the target size, raise an error.
        """
        raise NotImplementedError


from . import base as base_impl
from functools import partial
from xbarray.backends.pytorch import PytorchComputeBackend as BindingBackend

__all__ = [
    "quaternion_to_matrix",
    "matrix_to_quaternion",
    "euler_angles_to_matrix",
    "matrix_to_euler_angles",
    "random_quaternions",
    "random_rotations",
    "random_rotation",
    "standardize_quaternion",
    "quaternion_multiply",
    "quaternion_invert",
    "quaternion_apply",
    "axis_angle_to_matrix",
    "matrix_to_axis_angle",
    "axis_angle_to_quaternion",
    "quaternion_to_axis_angle",
    "rotation_6d_to_matrix",
    "matrix_to_rotation_6d",
]

quaternion_to_matrix = partial(base_impl.quaternion_to_matrix, BindingBackend)
matrix_to_quaternion = partial(base_impl.matrix_to_quaternion, BindingBackend)
euler_angles_to_matrix = partial(base_impl.euler_angles_to_matrix, BindingBackend)
matrix_to_euler_angles = partial(base_impl.matrix_to_euler_angles, BindingBackend)
random_quaternions = partial(base_impl.random_quaternions, BindingBackend)
random_rotations = partial(base_impl.random_rotations, BindingBackend)
random_rotation = partial(base_impl.random_rotation, BindingBackend)
standardize_quaternion = partial(base_impl.standardize_quaternion, BindingBackend)
quaternion_multiply = partial(base_impl.quaternion_multiply, BindingBackend)
quaternion_invert = partial(base_impl.quaternion_invert, BindingBackend)
quaternion_apply = partial(base_impl.quaternion_apply, BindingBackend)
axis_angle_to_matrix = partial(base_impl.axis_angle_to_matrix, BindingBackend)
matrix_to_axis_angle = partial(base_impl.matrix_to_axis_angle, BindingBackend)
axis_angle_to_quaternion = partial(base_impl.axis_angle_to_quaternion, BindingBackend)
quaternion_to_axis_angle = partial(base_impl.quaternion_to_axis_angle, BindingBackend)
rotation_6d_to_matrix = partial(base_impl.rotation_6d_to_matrix, BindingBackend)
matrix_to_rotation_6d = partial(base_impl.matrix_to_rotation_6d, BindingBackend)


from . import base as base_impl
from functools import partial
from xbarray.backends.pytorch import PytorchComputeBackend as BindingBackend

__all__ = [
    "pixel_coordinate_and_depth_to_world",
    "depth_image_to_world",
    "world_to_pixel_coordinate_and_depth",
    "world_to_depth",
]

pixel_coordinate_and_depth_to_world = partial(base_impl.pixel_coordinate_and_depth_to_world, BindingBackend)
depth_image_to_world = partial(base_impl.depth_image_to_world, BindingBackend)
world_to_pixel_coordinate_and_depth = partial(base_impl.world_to_pixel_coordinate_and_depth, BindingBackend)
world_to_depth = partial(base_impl.world_to_depth, BindingBackend)


from ._cls_base import ComputeBackendImplCls
from ._implementations import pytorch as pytorch_impl

class PytorchComputeBackend(metaclass=ComputeBackendImplCls[pytorch_impl.ARRAY_TYPE, pytorch_impl.DEVICE_TYPE, pytorch_impl.DTYPE_TYPE, pytorch_impl.RNG_TYPE]):
    ARRAY_TYPE = pytorch_impl.ARRAY_TYPE
    DTYPE_TYPE = pytorch_impl.DTYPE_TYPE
    DEVICE_TYPE = pytorch_impl.DEVICE_TYPE
    RNG_TYPE = pytorch_impl.RNG_TYPE

for name in dir(pytorch_impl):
    if not name.startswith('_') or name in [
        '__array_namespace_info__',
        '__array_api_version__',
    ]:
        try:
            setattr(PytorchComputeBackend, name, getattr(pytorch_impl, name))
        except AttributeError:
            pass

__all__ = [
    'PytorchComputeBackend',
]


from typing import Protocol, TypeVar, Optional, Any, Tuple, Union, Type, TypedDict, List, Literal, Sequence
from abc import abstractmethod
from ._array_typing import *
from ._api_return_typing import *

_NAMESPACE_ARRAY = TypeVar("_NAMESPACE_ARRAY", bound=Array)
_NAMESPACE_DEVICE = TypeVar("_NAMESPACE_DEVICE", bound=Device)
_NAMESPACE_DTYPE = TypeVar("_NAMESPACE_DTYPE", bound=DType)
class ArrayAPIFFTNamespace(Protocol[_NAMESPACE_ARRAY, _NAMESPACE_DEVICE, _NAMESPACE_DTYPE]):
    """
    FFT library API Stub
    https://github.com/data-apis/array-api/blob/main/src/array_api_stubs/_2024_12/fft.py
    """
    @abstractmethod
    def fft(
        self,
        x: _NAMESPACE_ARRAY,
        /,
        *,
        n: Optional[int] = None,
        axis: int = -1,
        norm: Literal["backward", "ortho", "forward"] = "backward",
    ) -> _NAMESPACE_ARRAY:
        """
        Computes the one-dimensional discrete Fourier transform.

        .. note::
        Applying the one-dimensional inverse discrete Fourier transform to the output of this function must return the original (i.e., non-transformed) input array within numerical accuracy (i.e., ``ifft(fft(x)) == x``), provided that the transform and inverse transform are performed with the same arguments (number of elements, axis, and normalization mode).

        Parameters
        ----------
        x: _NAMESPACE_ARRAY
            input array. Should have a complex floating-point data type.
        n: Optional[int]
            number of elements over which to compute the transform along the axis (dimension) specified by ``axis``. Let ``M`` be the size of the input array along the axis specified by ``axis``. When ``n`` is ``None``, the function must set ``n`` equal to ``M``.

            -   If ``n`` is greater than ``M``, the axis specified by ``axis`` must be zero-padded to size ``n``.
            -   If ``n`` is less than ``M``, the axis specified by ``axis`` must be trimmed to size ``n``.
            -   If ``n`` equals ``M``, all elements along the axis specified by ``axis`` must be used when computing the transform.

            Default: ``None``.
        axis: int
            axis (dimension) of the input array over which to compute the transform. A valid ``axis`` must be an integer on the interval ``[-N, N)``, where ``N`` is the rank (number of dimensions) of ``x``. If an ``axis`` is specified as a negative integer, the function must determine the axis along which to compute the transform by counting backward from the last dimension (where ``-1`` refers to the last dimension). Default: ``-1``.
        norm: Literal['backward', 'ortho', 'forward']
            normalization mode. Should be one of the following modes:

            - ``'backward'``: no normalization.
            - ``'ortho'``: normalize by ``1/sqrt(n)`` (i.e., make the FFT orthonormal).
            - ``'forward'``: normalize by ``1/n``.

            Default: ``'backward'``.

        Returns
        -------
        out: _NAMESPACE_ARRAY
            an array transformed along the axis (dimension) specified by ``axis``. The returned array must have the same data type as ``x`` and must have the same shape as ``x``, except for the axis specified by ``axis`` which must have size ``n``.

        Notes
        -----

        .. versionadded:: 2022.12

        .. versionchanged:: 2023.12
        Required the input array have a complex floating-point data type and required that the output array have the same data type as the input array.
        """

    @abstractmethod
    def ifft(
        self,
        x: _NAMESPACE_ARRAY,
        /,
        *,
        n: Optional[int] = None,
        axis: int = -1,
        norm: Literal["backward", "ortho", "forward"] = "backward",
    ) -> _NAMESPACE_ARRAY:
        """
        Computes the one-dimensional inverse discrete Fourier transform.

        .. note::
        Applying the one-dimensional inverse discrete Fourier transform to the output of this function must return the original (i.e., non-transformed) input array within numerical accuracy (i.e., ``ifft(fft(x)) == x``), provided that the transform and inverse transform are performed with the same arguments (number of elements, axis, and normalization mode).

        Parameters
        ----------
        x: _NAMESPACE_ARRAY
            input array. Should have a complex floating-point data type.
        n: Optional[int]
            number of elements over which to compute the transform along the axis (dimension) specified by ``axis``. Let ``M`` be the size of the input array along the axis specified by ``axis``. When ``n`` is ``None``, the function must set ``n`` equal to ``M``.

            -   If ``n`` is greater than ``M``, the axis specified by ``axis`` must be zero-padded to size ``n``.
            -   If ``n`` is less than ``M``, the axis specified by ``axis`` must be trimmed to size ``n``.
            -   If ``n`` equals ``M``, all elements along the axis specified by ``axis`` must be used when computing the transform.

            Default: ``None``.
        axis: int
            axis (dimension) of the input array over which to compute the transform. A valid ``axis`` must be an integer on the interval ``[-N, N)``, where ``N`` is the rank (number of dimensions) of ``x``. If an ``axis`` is specified as a negative integer, the function must determine the axis along which to compute the transform by counting backward from the last dimension (where ``-1`` refers to the last dimension). Default: ``-1``.
        norm: Literal['backward', 'ortho', 'forward']
            normalization mode. Should be one of the following modes:

            - ``'backward'``: normalize by ``1/n``.
            - ``'ortho'``: normalize by ``1/sqrt(n)`` (i.e., make the FFT orthonormal).
            - ``'forward'``: no normalization.

            Default: ``'backward'``.

        Returns
        -------
        out: _NAMESPACE_ARRAY
            an array transformed along the axis (dimension) specified by ``axis``. The returned array must have the same data type as ``x`` and must have the same shape as ``x``, except for the axis specified by ``axis`` which must have size ``n``.

        Notes
        -----

        .. versionadded:: 2022.12

        .. versionchanged:: 2023.12
        Required the input array have a complex floating-point data type and required that the output array have the same data type as the input array.
        """

    @abstractmethod
    def fftn(
        self,
        x: _NAMESPACE_ARRAY,
        /,
        *,
        s: Optional[Sequence[int]] = None,
        axes: Optional[Sequence[int]] = None,
        norm: Literal["backward", "ortho", "forward"] = "backward",
    ) -> _NAMESPACE_ARRAY:
        """
        Computes the n-dimensional discrete Fourier transform.

        .. note::
        Applying the n-dimensional inverse discrete Fourier transform to the output of this function must return the original (i.e., non-transformed) input array within numerical accuracy (i.e., ``ifftn(fftn(x)) == x``), provided that the transform and inverse transform are performed with the same arguments (sizes, axes, and normalization mode).

        Parameters
        ----------
        x: _NAMESPACE_ARRAY
            input array. Should have a complex floating-point data type.
        s: Optional[Sequence[int]]
            number of elements over which to compute the transform along the axes (dimensions) specified by ``axes``. Let ``i`` be the index of the ``n``-th axis specified by ``axes`` (i.e., ``i = axes[n]``) and ``M[i]`` be the size of the input array along axis ``i``. When ``s`` is ``None``, the function must set ``s`` equal to a sequence of integers such that ``s[i]`` equals ``M[i]`` for all ``i``.

            -   If ``s[i]`` is greater than ``M[i]``, axis ``i`` must be zero-padded to size ``s[i]``.
            -   If ``s[i]`` is less than ``M[i]``, axis ``i`` must be trimmed to size ``s[i]``.
            -   If ``s[i]`` equals ``M[i]`` or ``-1``, all elements along axis ``i`` must be used when computing the transform.

            If ``s`` is not ``None``, ``axes`` must not be ``None``. Default: ``None``.
        axes: Optional[Sequence[int]]
            axes (dimensions) over which to compute the transform. A valid axis in ``axes`` must be an integer on the interval ``[-N, N)``, where ``N`` is the rank (number of dimensions) of ``x``. If an axis is specified as a negative integer, the function must determine the axis along which to compute the transform by counting backward from the last dimension (where ``-1`` refers to the last dimension).

            If ``s`` is provided, the corresponding ``axes`` to be transformed must also be provided. If ``axes`` is ``None``, the function must compute the transform over all axes. Default: ``None``.

            If ``axes`` contains two or more entries which resolve to the same axis (i.e., resolved axes are not unique), the behavior is unspecified and thus implementation-defined.
        norm: Literal['backward', 'ortho', 'forward']
            normalization mode. Should be one of the following modes:

            - ``'backward'``: no normalization.
            - ``'ortho'``: normalize by ``1/sqrt(n)`` (i.e., make the FFT orthonormal).
            - ``'forward'``: normalize by ``1/n``.

            where ``n = prod(s)`` is the logical FFT size.

            Default: ``'backward'``.

        Returns
        -------
        out: _NAMESPACE_ARRAY
            an array transformed along the axes (dimensions) specified by ``axes``. The returned array must have the same data type as ``x`` and must have the same shape as ``x``, except for the axes specified by ``axes`` which must have size ``s[i]``.

        Notes
        -----

        .. versionadded:: 2022.12

        .. versionchanged:: 2023.12
        Required the input array have a complex floating-point data type and required that the output array have the same data type as the input array.
        """

    @abstractmethod
    def ifftn(
        self,
        x: _NAMESPACE_ARRAY,
        /,
        *,
        s: Optional[Sequence[int]] = None,
        axes: Optional[Sequence[int]] = None,
        norm: Literal["backward", "ortho", "forward"] = "backward",
    ) -> _NAMESPACE_ARRAY:
        """
        Computes the n-dimensional inverse discrete Fourier transform.

        .. note::
        Applying the n-dimensional inverse discrete Fourier transform to the output of this function must return the original (i.e., non-transformed) input array within numerical accuracy (i.e., ``ifftn(fftn(x)) == x``), provided that the transform and inverse transform are performed with the same arguments (sizes, axes, and normalization mode).

        Parameters
        ----------
        x: _NAMESPACE_ARRAY
            input array. Should have a complex floating-point data type.
        s: Optional[Sequence[int]]
            number of elements over which to compute the transform along the axes (dimensions) specified by ``axes``. Let ``i`` be the index of the ``n``-th axis specified by ``axes`` (i.e., ``i = axes[n]``) and ``M[i]`` be the size of the input array along axis ``i``. When ``s`` is ``None``, the function must set ``s`` equal to a sequence of integers such that ``s[i]`` equals ``M[i]`` for all ``i``.

            -   If ``s[i]`` is greater than ``M[i]``, axis ``i`` must be zero-padded to size ``s[i]``.
            -   If ``s[i]`` is less than ``M[i]``, axis ``i`` must be trimmed to size ``s[i]``.
            -   If ``s[i]`` equals ``M[i]`` or ``-1``, all elements along axis ``i`` must be used when computing the transform.

            If ``s`` is not ``None``, ``axes`` must not be ``None``. Default: ``None``.
        axes: Optional[Sequence[int]]
            axes (dimensions) over which to compute the transform. A valid axis in ``axes`` must be an integer on the interval ``[-N, N)``, where ``N`` is the rank (number of dimensions) of ``x``. If an axis is specified as a negative integer, the function must determine the axis along which to compute the transform by counting backward from the last dimension (where ``-1`` refers to the last dimension).

            If ``s`` is provided, the corresponding ``axes`` to be transformed must also be provided. If ``axes`` is ``None``, the function must compute the transform over all axes. Default: ``None``.

            If ``axes`` contains two or more entries which resolve to the same axis (i.e., resolved axes are not unique), the behavior is unspecified and thus implementation-defined.
        norm: Literal['backward', 'ortho', 'forward']
            specify the normalization mode. Should be one of the following modes:

            - ``'backward'``: normalize by ``1/n``.
            - ``'ortho'``: normalize by ``1/sqrt(n)`` (i.e., make the FFT orthonormal).
            - ``'forward'``: no normalization.

            where ``n = prod(s)`` is the logical FFT size.

            Default: ``'backward'``.

        Returns
        -------
        out: _NAMESPACE_ARRAY
            an array transformed along the axes (dimensions) specified by ``axes``. The returned array must have the same data type as ``x`` and must have the same shape as ``x``, except for the axes specified by ``axes`` which must have size ``s[i]``.

        Notes
        -----

        .. versionadded:: 2022.12

        .. versionchanged:: 2023.12
        Required the input array have a complex floating-point data type and required that the output array have the same data type as the input array.
        """

    @abstractmethod
    def rfft(
        self,
        x: _NAMESPACE_ARRAY,
        /,
        *,
        n: Optional[int] = None,
        axis: int = -1,
        norm: Literal["backward", "ortho", "forward"] = "backward",
    ) -> _NAMESPACE_ARRAY:
        """
        Computes the one-dimensional discrete Fourier transform for real-valued input.

        .. note::
        Applying the one-dimensional inverse discrete Fourier transform for real-valued input to the output of this function must return the original (i.e., non-transformed) input array within numerical accuracy (i.e., ``irfft(rfft(x)) == x``), provided that the transform and inverse transform are performed with the same arguments (axis and normalization mode) and consistent values for the number of elements over which to compute the transforms.

        Parameters
        ----------
        x: _NAMESPACE_ARRAY
            input array. Must have a real-valued floating-point data type.
        n: Optional[int]
            number of elements over which to compute the transform along the axis (dimension) specified by ``axis``. Let ``M`` be the size of the input array along the axis specified by ``axis``. When ``n`` is ``None``, the function must set ``n`` equal to ``M``.

            -   If ``n`` is greater than ``M``, the axis specified by ``axis`` must be zero-padded to size ``n``.
            -   If ``n`` is less than ``M``, the axis specified by ``axis`` must be trimmed to size ``n``.
            -   If ``n`` equals ``M``, all elements along the axis specified by ``axis`` must be used when computing the transform.

            Default: ``None``.
        axis: int
            axis (dimension) of the input array over which to compute the transform. A valid ``axis`` must be an integer on the interval ``[-N, N)``, where ``N`` is the rank (number of dimensions) of ``x``. If an ``axis`` is specified as a negative integer, the function must determine the axis along which to compute the transform by counting backward from the last dimension (where ``-1`` refers to the last dimension). Default: ``-1``.
        norm: Literal['backward', 'ortho', 'forward']
            normalization mode. Should be one of the following modes:

            - ``'backward'``: no normalization.
            - ``'ortho'``: normalize by ``1/sqrt(n)`` (i.e., make the FFT orthonormal).
            - ``'forward'``: normalize by ``1/n``.

            Default: ``'backward'``.

        Returns
        -------
        out: _NAMESPACE_ARRAY
            an array transformed along the axis (dimension) specified by ``axis``. The returned array must have a complex floating-point data type whose precision matches the precision of ``x`` (e.g., if ``x`` is ``float64``, then the returned array must have a ``complex128`` data type). The returned array must have the same shape as ``x``, except for the axis specified by ``axis`` which must have size ``n//2 + 1``.

        Notes
        -----

        .. versionadded:: 2022.12
        """

    @abstractmethod
    def irfft(
        self,
        x: _NAMESPACE_ARRAY,
        /,
        *,
        n: Optional[int] = None,
        axis: int = -1,
        norm: Literal["backward", "ortho", "forward"] = "backward",
    ) -> _NAMESPACE_ARRAY:
        """
        Computes the one-dimensional inverse of ``rfft`` for complex-valued input.

        .. note::
        Applying the one-dimensional inverse discrete Fourier transform for real-valued input to the output of this function must return the original (i.e., non-transformed) input array within numerical accuracy (i.e., ``irfft(rfft(x)) == x``), provided that the transform and inverse transform are performed with the same arguments (axis and normalization mode) and consistent values for the number of elements over which to compute the transforms.

        Parameters
        ----------
        x: _NAMESPACE_ARRAY
            input array. Should have a complex floating-point data type.
        n: Optional[int]
            number of elements along the transformed axis (dimension) specified by ``axis`` in the **output array**. Let ``M`` be the size of the input array along the axis specified by ``axis``. When ``n`` is ``None``, the function must set ``n`` equal to ``2*(M-1)``.

            -   If ``n//2+1`` is greater than ``M``, the axis of the input array specified by ``axis`` must be zero-padded to size ``n//2+1``.
            -   If ``n//2+1`` is less than ``M``, the axis of the input array specified by ``axis`` must be trimmed to size ``n//2+1``.
            -   If ``n//2+1`` equals ``M``, all elements along the axis of the input array specified by ``axis`` must be used when computing the transform.

            Default: ``None``.
        axis: int
            axis (dimension) of the input array over which to compute the transform. A valid ``axis`` must be an integer on the interval ``[-N, N)``, where ``N`` is the rank (number of dimensions) of ``x``. If an ``axis`` is specified as a negative integer, the function must determine the axis along which to compute the transform by counting backward from the last dimension (where ``-1`` refers to the last dimension). Default: ``-1``.
        norm: Literal['backward', 'ortho', 'forward']
            normalization mode. Should be one of the following modes:

            - ``'backward'``: normalize by ``1/n``.
            - ``'ortho'``: normalize by ``1/sqrt(n)`` (i.e., make the FFT orthonormal).
            - ``'forward'``: no normalization.

            Default: ``'backward'``.

        Returns
        -------
        out: _NAMESPACE_ARRAY
            an array transformed along the axis (dimension) specified by ``axis``. The returned array must have a real-valued floating-point data type whose precision matches the precision of ``x`` (e.g., if ``x`` is ``complex128``, then the returned array must have a ``float64`` data type). The returned array must have the same shape as ``x``, except for the axis specified by ``axis`` which must have size ``n``.

        Notes
        -----

        -   In order to return an array having an odd number of elements along the transformed axis, the function must be provided an odd integer for ``n``.

        .. versionadded:: 2022.12

        .. versionchanged:: 2023.12
        Required the output array have a real-valued floating-point data type having the same precision as the input array.
        """

    @abstractmethod
    def rfftn(
        self,
        x: _NAMESPACE_ARRAY,
        /,
        *,
        s: Optional[Sequence[int]] = None,
        axes: Optional[Sequence[int]] = None,
        norm: Literal["backward", "ortho", "forward"] = "backward",
    ) -> _NAMESPACE_ARRAY:
        """
        Computes the n-dimensional discrete Fourier transform for real-valued input.

        .. note::
        Applying the n-dimensional inverse discrete Fourier transform for real-valued input to the output of this function must return the original (i.e., non-transformed) input array within numerical accuracy (i.e., ``irfftn(rfftn(x)) == x``), provided that the transform and inverse transform are performed with the same arguments (axes and normalization mode) and consistent sizes.

        Parameters
        ----------
        x: _NAMESPACE_ARRAY
            input array. Must have a real-valued floating-point data type.
        s: Optional[Sequence[int]]
            number of elements over which to compute the transform along axes (dimensions) specified by ``axes``. Let ``i`` be the index of the ``n``-th axis specified by ``axes`` (i.e., ``i = axes[n]``) and ``M[i]`` be the size of the input array along axis ``i``. When ``s`` is ``None``, the function must set ``s`` equal to a sequence of integers such that ``s[i]`` equals ``M[i]`` for all ``i``.

            -   If ``s[i]`` is greater than ``M[i]``, axis ``i`` must be zero-padded to size ``s[i]``.
            -   If ``s[i]`` is less than ``M[i]``, axis ``i`` must be trimmed to size ``s[i]``.
            -   If ``s[i]`` equals ``M[i]`` or ``-1``, all elements along axis ``i`` must be used when computing the transform.

            If ``s`` is not ``None``, ``axes`` must not be ``None``. Default: ``None``.
        axes: Optional[Sequence[int]]
            axes (dimensions) over which to compute the transform. A valid axis in ``axes`` must be an integer on the interval ``[-N, N)``, where ``N`` is the rank (number of dimensions) of ``x``. If an axis is specified as a negative integer, the function must determine the axis along which to compute the transform by counting backward from the last dimension (where ``-1`` refers to the last dimension).

            If ``s`` is provided, the corresponding ``axes`` to be transformed must also be provided. If ``axes`` is ``None``, the function must compute the transform over all axes. Default: ``None``.

            If ``axes`` contains two or more entries which resolve to the same axis (i.e., resolved axes are not unique), the behavior is unspecified and thus implementation-defined.
        norm: Literal['backward', 'ortho', 'forward']
            normalization mode. Should be one of the following modes:

            - ``'backward'``: no normalization.
            - ``'ortho'``: normalize by ``1/sqrt(n)`` (i.e., make the FFT orthonormal).
            - ``'forward'``: normalize by ``1/n``.

            where ``n = prod(s)``, the logical FFT size.

            Default: ``'backward'``.

        Returns
        -------
        out: _NAMESPACE_ARRAY
            an array transformed along the axes (dimension) specified by ``axes``. The returned array must have a complex floating-point data type whose precision matches the precision of ``x`` (e.g., if ``x`` is ``float64``, then the returned array must have a ``complex128`` data type). The returned array must have the same shape as ``x``, except for the last transformed axis which must have size ``s[-1]//2 + 1`` and the remaining transformed axes which must have size ``s[i]``.

        Notes
        -----

        .. versionadded:: 2022.12
        """

    @abstractmethod
    def irfftn(
        self,
        x: _NAMESPACE_ARRAY,
        /,
        *,
        s: Optional[Sequence[int]] = None,
        axes: Optional[Sequence[int]] = None,
        norm: Literal["backward", "ortho", "forward"] = "backward",
    ) -> _NAMESPACE_ARRAY:
        """
        Computes the n-dimensional inverse of ``rfftn`` for complex-valued input.

        .. note::
        Applying the n-dimensional inverse discrete Fourier transform for real-valued input to the output of this function must return the original (i.e., non-transformed) input array within numerical accuracy (i.e., ``irfftn(rfftn(x)) == x``), provided that the transform and inverse transform are performed with the same arguments (axes and normalization mode) and consistent sizes.

        Parameters
        ----------
        x: _NAMESPACE_ARRAY
            input array. Should have a complex floating-point data type.
        s: Optional[Sequence[int]]
            number of elements along the transformed axes (dimensions) specified by ``axes`` in the **output array**. Let ``i`` be the index of the ``n``-th axis specified by ``axes`` (i.e., ``i = axes[n]``) and ``M[i]`` be the size of the input array along axis ``i``. When ``s`` is ``None``, the function must set ``s`` equal to a sequence of integers such that ``s[i]`` equals ``M[i]`` for all ``i``, except for the last transformed axis in which ``s[i]`` equals ``2*(M[i]-1)``. For each ``i``, let ``n`` equal ``s[i]``, except for the last transformed axis in which ``n`` equals ``s[i]//2+1``.

            -   If ``n`` is greater than ``M[i]``, axis ``i`` of the input array must be zero-padded to size ``n``.
            -   If ``n`` is less than ``M[i]``, axis ``i`` of the input array must be trimmed to size ``n``.
            -   If ``n`` equals ``M[i]`` or ``-1``, all elements along axis ``i`` of the input array must be used when computing the transform.

            If ``s`` is not ``None``, ``axes`` must not be ``None``. Default: ``None``.
        axes: Optional[Sequence[int]]
            axes (dimensions) over which to compute the transform. A valid axis in ``axes`` must be an integer on the interval ``[-N, N)``, where ``N`` is the rank (number of dimensions) of ``x``. If an axis is specified as a negative integer, the function must determine the axis along which to compute the transform by counting backward from the last dimension (where ``-1`` refers to the last dimension).

            If ``s`` is provided, the corresponding ``axes`` to be transformed must also be provided. If ``axes`` is ``None``, the function must compute the transform over all axes. Default: ``None``.

            If ``axes`` contains two or more entries which resolve to the same axis (i.e., resolved axes are not unique), the behavior is unspecified and thus implementation-defined.
        norm: Literal['backward', 'ortho', 'forward']
            normalization mode. Should be one of the following modes:

            - ``'backward'``: normalize by ``1/n``.
            - ``'ortho'``: normalize by ``1/sqrt(n)`` (i.e., make the FFT orthonormal).
            - ``'forward'``: no normalization.

            where ``n = prod(s)`` is the logical FFT size.

            Default: ``'backward'``.

        Returns
        -------
        out: _NAMESPACE_ARRAY
            an array transformed along the axes (dimension) specified by ``axes``. The returned array must have a real-valued floating-point data type whose precision matches the precision of ``x`` (e.g., if ``x`` is ``complex128``, then the returned array must have a ``float64`` data type). The returned array must have the same shape as ``x``, except for the transformed axes which must have size ``s[i]``.

        Notes
        -----

        -   In order to return an array having an odd number of elements along the last transformed axis, the function must be provided an odd integer for ``s[-1]``.

        .. versionadded:: 2022.12

        .. versionchanged:: 2023.12
        Required the output array have a real-valued floating-point data type having the same precision as the input array.
        """

    @abstractmethod
    def hfft(
        self,
        x: _NAMESPACE_ARRAY,
        /,
        *,
        n: Optional[int] = None,
        axis: int = -1,
        norm: Literal["backward", "ortho", "forward"] = "backward",
    ) -> _NAMESPACE_ARRAY:
        """
        Computes the one-dimensional discrete Fourier transform of a signal with Hermitian symmetry.

        Parameters
        ----------
        x: _NAMESPACE_ARRAY
            input array. Should have a complex floating-point data type.
        n: Optional[int]
            number of elements along the transformed axis (dimension) specified by ``axis`` in the **output array**. Let ``M`` be the size of the input array along the axis specified by ``axis``. When ``n`` is ``None``, the function must set ``n`` equal to ``2*(M-1)``.

            -   If ``n//2+1`` is greater than ``M``, the axis of the input array specified by ``axis`` must be zero-padded to length ``n//2+1``.
            -   If ``n//2+1`` is less than ``M``, the axis of the input array specified by ``axis`` must be trimmed to size ``n//2+1``.
            -   If ``n//2+1`` equals ``M``, all elements along the axis of the input array specified by ``axis`` must be used when computing the transform.

            Default: ``None``.
        axis: int
            axis (dimension) of the input array over which to compute the transform. A valid ``axis`` must be an integer on the interval ``[-N, N)``, where ``N`` is the rank (number of dimensions) of ``x``. If an ``axis`` is specified as a negative integer, the function must determine the axis along which to compute the transform by counting backward from the last dimension (where ``-1`` refers to the last dimension). Default: ``-1``.
        norm: Literal['backward', 'ortho', 'forward']
            normalization mode. Should be one of the following modes:

            - ``'backward'``: no normalization.
            - ``'ortho'``: normalize by ``1/sqrt(n)`` (i.e., make the FFT orthonormal).
            - ``'forward'``: normalize by ``1/n``.

            Default: ``'backward'``.

        Returns
        -------
        out: _NAMESPACE_ARRAY
            an array transformed along the axis (dimension) specified by ``axis``. The returned array must have a real-valued floating-point data type whose precision matches the precision of ``x`` (e.g., if ``x`` is ``complex128``, then the returned array must have a ``float64`` data type). The returned array must have the same shape as ``x``, except for the axis specified by ``axis`` which must have size ``n``.

        Notes
        -----

        .. versionadded:: 2022.12

        .. versionchanged:: 2023.12
        Required the input array to have a complex floating-point data type and required that the output array have a real-valued data type having the same precision as the input array.
        """

    @abstractmethod
    def ihfft(
        self,
        x: _NAMESPACE_ARRAY,
        /,
        *,
        n: Optional[int] = None,
        axis: int = -1,
        norm: Literal["backward", "ortho", "forward"] = "backward",
    ) -> _NAMESPACE_ARRAY:
        """
        Computes the one-dimensional inverse discrete Fourier transform of a signal with Hermitian symmetry.

        Parameters
        ----------
        x: _NAMESPACE_ARRAY
            input array. Must have a real-valued floating-point data type.
        n: Optional[int]
            number of elements over which to compute the transform along the axis (dimension) specified by ``axis``. Let ``M`` be the size of the input array along the axis specified by ``axis``. When ``n`` is ``None``, the function must set ``n`` equal to ``M``.

            -   If ``n`` is greater than ``M``, the axis specified by ``axis`` must be zero-padded to size ``n``.
            -   If ``n`` is less than ``M``, the axis specified by ``axis`` must be trimmed to size ``n``.
            -   If ``n`` equals ``M``, all elements along the axis specified by ``axis`` must be used when computing the transform.

            Default: ``None``.
        axis: int
            axis (dimension) of the input array over which to compute the transform. A valid ``axis`` must be an integer on the interval ``[-N, N)``, where ``N`` is the rank (number of dimensions) of ``x``. If an ``axis`` is specified as a negative integer, the function must determine the axis along which to compute the transform by counting backward from the last dimension (where ``-1`` refers to the last dimension). Default: ``-1``.
        norm: Literal['backward', 'ortho', 'forward']
            normalization mode. Should be one of the following modes:

            - ``'backward'``: normalize by ``1/n``.
            - ``'ortho'``: normalize by ``1/sqrt(n)`` (i.e., make the FFT orthonormal).
            - ``'forward'``: no normalization.

            Default: ``'backward'``.

        Returns
        -------
        out: _NAMESPACE_ARRAY
            an array transformed along the axis (dimension) specified by ``axis``. The returned array must have a complex floating-point data type whose precision matches the precision of ``x`` (e.g., if ``x`` is ``float64``, then the returned array must have a ``complex128`` data type). The returned array must have the same shape as ``x``, except for the axis specified by ``axis`` which must have size ``n//2 + 1``.

        Notes
        -----

        .. versionadded:: 2022.12
        """

    @abstractmethod
    def fftfreq(
        self,
        n: int,
        /,
        *,
        d: float = 1.0,
        dtype: Optional[_NAMESPACE_DTYPE] = None,
        device: Optional[_NAMESPACE_DEVICE] = None,
    ) -> _NAMESPACE_ARRAY:
        """
        Computes the discrete Fourier transform sample frequencies.

        For a Fourier transform of length ``n`` and length unit of ``d``, the frequencies are described as:

        .. code-block::

        f = [0, 1, ..., n/2-1, -n/2, ..., -1] / (d*n)        # if n is even
        f = [0, 1, ..., (n-1)/2, -(n-1)/2, ..., -1] / (d*n)  # if n is odd

        Parameters
        ----------
        n: int
            window length.
        d: float
            sample spacing between individual samples of the Fourier transform input. Default: ``1.0``.
        dtype: Optional[dtype]
            output array data type. Must be a real-valued floating-point data type. If ``dtype`` is ``None``, the output array data type must be the default real-valued floating-point data type. Default: ``None``.
        device: Optional[device]
            device on which to place the created array. Default: ``None``.

        Returns
        -------
        out: _NAMESPACE_ARRAY
            an array of shape ``(n,)`` containing the sample frequencies.

        Notes
        -----

        .. versionadded:: 2022.12

        .. versionchanged:: 2023.12
        Required the output array have the default real-valued floating-point data type.

        .. versionchanged:: 2024.12
        Added ``dtype`` keyword argument support.
        """

    @abstractmethod
    def rfftfreq(
        self,
        n: int,
        /,
        *,
        d: float = 1.0,
        dtype: Optional[_NAMESPACE_DTYPE] = None,
        device: Optional[_NAMESPACE_DEVICE] = None,
    ) -> _NAMESPACE_ARRAY:
        """
        Computes the discrete Fourier transform sample frequencies (for ``rfft`` and ``irfft``).

        For a Fourier transform of length ``n`` and length unit of ``d``, the frequencies are described as:

        .. code-block::

        f = [0, 1, ...,     n/2-1,     n/2] / (d*n)  # if n is even
        f = [0, 1, ..., (n-1)/2-1, (n-1)/2] / (d*n)  # if n is odd

        The Nyquist frequency component is considered to be positive.

        Parameters
        ----------
        n: int
            window length.
        d: float
            sample spacing between individual samples of the Fourier transform input. Default: ``1.0``.
        dtype: Optional[dtype]
            output array data type. Must be a real-valued floating-point data type. If ``dtype`` is ``None``, the output array data type must be the default real-valued floating-point data type. Default: ``None``.
        device: Optional[device]
            device on which to place the created array. Default: ``None``.

        Returns
        -------
        out: _NAMESPACE_ARRAY
            an array of shape ``(n//2+1,)`` containing the sample frequencies.

        Notes
        -----

        .. versionadded:: 2022.12

        .. versionchanged:: 2023.12
        Required the output array have the default real-valued floating-point data type.

        .. versionchanged:: 2024.12
        Added ``dtype`` keyword argument support.
        """

    @abstractmethod
    def fftshift(
        self,
        x: _NAMESPACE_ARRAY, /, *, axes: Optional[Union[int, Sequence[int]]] = None
    ) -> _NAMESPACE_ARRAY:
        """
        Shifts the zero-frequency component to the center of the spectrum.

        This function swaps half-spaces for all axes (dimensions) specified by ``axes``.

        .. note::
        ``out[0]`` is the Nyquist component only if the length of the input is even.

        Parameters
        ----------
        x: _NAMESPACE_ARRAY
            input array. Should have a floating-point data type.
        axes: Optional[Union[int, Sequence[int]]]
            axes over which to shift. If ``None``, the function must shift all axes. Default: ``None``.

            If ``axes`` contains two or more entries which resolve to the same axis (i.e., resolved axes are not unique), the behavior is unspecified and thus implementation-defined.

        Returns
        -------
        out: _NAMESPACE_ARRAY
            the shifted array. The returned array must have the same data type and shape as ``x``.

        Notes
        -----

        .. versionadded:: 2022.12
        """

    @abstractmethod
    def ifftshift(
        self,
        x: _NAMESPACE_ARRAY, /, *, axes: Optional[Union[int, Sequence[int]]] = None
    ) -> _NAMESPACE_ARRAY:
        """
        Inverse of ``fftshift``.

        .. note::
        Although identical for even-length ``x``, ``fftshift`` and ``ifftshift`` differ by one sample for odd-length ``x``.

        Parameters
        ----------
        x: _NAMESPACE_ARRAY
            input array. Should have a floating-point data type.
        axes: Optional[Union[int, Sequence[int]]]
            axes over which to perform the inverse shift. If ``None``, the function must shift all axes. Default: ``None``.

            If ``axes`` contains two or more entries which resolve to the same axis (i.e., resolved axes are not unique), the behavior is unspecified and thus implementation-defined.

        Returns
        -------
        out: _NAMESPACE_ARRAY
            the shifted array. The returned array must have the same data type and shape as ``x``.

        Notes
        -----

        .. versionadded:: 2022.12
        """


"""
From https://github.com/data-apis/array-api/blob/main/src/array_api_stubs/_2024_12/constants.py
"""

e = 2.718281828459045
"""
IEEE 754 floating-point representation of Euler's constant.

``e = 2.71828182845904523536028747135266249775724709369995...``
"""

inf = float("inf")
"""
IEEE 754 floating-point representation of (positive) infinity.
"""

nan = float("nan")
"""
IEEE 754 floating-point representation of Not a Number (``NaN``).
"""

newaxis = None
"""
An alias for ``None`` which is useful for indexing arrays.
"""

pi = 3.141592653589793
"""
IEEE 754 floating-point representation of the mathematical constant ``π``.

``pi = 3.1415926535897932384626433...``
"""


from ._cls_base import ComputeBackendImplCls
from ._implementations import numpy as numpy_impl

class NumpyComputeBackend(metaclass=ComputeBackendImplCls[numpy_impl.ARRAY_TYPE, numpy_impl.DEVICE_TYPE, numpy_impl.DTYPE_TYPE, numpy_impl.RNG_TYPE]):
    ARRAY_TYPE = numpy_impl.ARRAY_TYPE
    DTYPE_TYPE = numpy_impl.DTYPE_TYPE
    DEVICE_TYPE = numpy_impl.DEVICE_TYPE
    RNG_TYPE = numpy_impl.RNG_TYPE

for name in dir(numpy_impl):
    if not name.startswith('_') or name in [
        '__array_namespace_info__',
        '__array_api_version__',
    ]:
        setattr(NumpyComputeBackend, name, getattr(numpy_impl, name))

__all__ = [
    'NumpyComputeBackend',
]


from typing import Any, Union, Optional
import numpy as np
import torch
from ._typing import ARRAY_TYPE, DTYPE_TYPE, DEVICE_TYPE, RNG_TYPE
from xbarray.backends.base import ComputeBackend, SupportsDLPack

PYTORCH_DTYPE_CAST_MAP = {
    torch.uint16: torch.int16,
    torch.uint32: torch.int32,
    torch.uint64: torch.int64,
    torch.float8_e4m3fn: torch.float16,
    torch.float8_e5m2: torch.float16,
}

__all__ = [
    "default_integer_dtype",
    "default_floating_dtype",
    "default_boolean_dtype",
    "is_backendarray",
    "from_numpy",
    "from_other_backend",
    "to_numpy",
    "to_dlpack",
    "dtype_is_real_integer",
    "dtype_is_real_floating",
    "dtype_is_boolean",
    "abbreviate_array",
    "map_fn_over_arrays",
]

default_integer_dtype = torch.int32
default_floating_dtype = torch.float32
default_boolean_dtype = torch.bool

def is_backendarray(data : Any) -> bool:
    return isinstance(data, torch.Tensor)

def from_numpy(
    data : np.ndarray,
    /,
    *,
    dtype : Optional[DTYPE_TYPE] = None,
    device : Optional[DEVICE_TYPE] = None
) -> ARRAY_TYPE:
    t = torch.from_numpy(data)
    target_dtype = dtype if dtype is not None else PYTORCH_DTYPE_CAST_MAP.get(t.dtype, t.dtype)
    if target_dtype is not None or device is not None:
        t = t.to(device=device, dtype=target_dtype)
    return t

def from_other_backend(
    other_backend: ComputeBackend,
    data: Any,
    /,
) -> ARRAY_TYPE:
    dat_dlpack = other_backend.to_dlpack(data)
    return torch.from_dlpack(dat_dlpack)

def to_numpy(
    data : ARRAY_TYPE
) -> np.ndarray:
    # Torch bfloat16 is not supported by numpy
    if data.dtype == torch.bfloat16:
        data = data.to(torch.float32)
    return data.cpu().numpy()

def to_dlpack(
    data: ARRAY_TYPE,
    /,
) -> SupportsDLPack:
    return data

def dtype_is_real_integer(
    dtype: DTYPE_TYPE
) -> bool:
    # https://pytorch.org/docs/stable/tensors.html#id12
    return dtype in [
        torch.int8, torch.int16, torch.int32, torch.int64,
        torch.uint8, 
        torch.int,
        torch.long
    ]

def dtype_is_real_floating(
    dtype: DTYPE_TYPE
) -> bool:
    return dtype in [
        torch.float16, torch.float32, torch.float64, 
        torch.float, torch.double, 
        torch.bfloat16
    ]

def dtype_is_boolean(
    dtype: DTYPE_TYPE
) -> bool:
    return dtype == torch.bool

from .._common.implementations import *
from array_api_compat import torch as compat_module
abbreviate_array = get_abbreviate_array_function(
    compat_module, 
    default_integer_dtype=default_integer_dtype, 
    func_dtype_is_real_floating=dtype_is_real_floating,
    func_dtype_is_real_integer=dtype_is_real_integer,
    func_dtype_is_boolean=dtype_is_boolean
)
map_fn_over_arrays = get_map_fn_over_arrays_function(
    is_backendarray=is_backendarray,
)
pad_dim = get_pad_dim_function(
    backend=compat_module,
)


from typing import Any, Union, Optional, Callable
import jax
import jax.numpy as jnp
import numpy as np
from ._typing import ARRAY_TYPE, DTYPE_TYPE, DEVICE_TYPE, RNG_TYPE
from xbarray.backends.base import ComputeBackend, SupportsDLPack

__all__ = [
    "default_integer_dtype",
    "default_floating_dtype",
    "default_boolean_dtype",
    "is_backendarray",
    "from_numpy",
    "from_other_backend",
    "to_numpy",
    "to_dlpack",
    "dtype_is_real_integer",
    "dtype_is_real_floating",
    "dtype_is_boolean",
    "abbreviate_array",
    "map_fn_over_arrays",
]

default_integer_dtype = int
default_floating_dtype = float
default_boolean_dtype = bool

def is_backendarray(data : Any) -> bool:
    return isinstance(data, jax.Array)

def from_numpy(
    data : np.ndarray,
    /,
    *,
    dtype : Optional[DTYPE_TYPE] = None,
    device : Optional[DEVICE_TYPE] = None
) -> ARRAY_TYPE:
    return jax.numpy.asarray(data, dtype=dtype, device=device)

def from_other_backend(
    other_backend: ComputeBackend,
    data: Any,
    /,
) -> ARRAY_TYPE:
    data_dlpack = other_backend.to_dlpack(data)
    return jax.dlpack.from_dlpack(data_dlpack)
    # except Exception as e:
    #     # jax sometimes has tiling issues with dlpack converted data
    #     np = other_backend.to_numpy(data)
    #     return from_numpy(np)

def to_numpy(
    data : ARRAY_TYPE
) -> np.ndarray:
    if data.dtype == jax.dtypes.bfloat16:
        data = data.astype(np.float32)
    return np.asarray(data)

def to_dlpack(
    data: ARRAY_TYPE,
    /,
) -> SupportsDLPack:
    return data

def dtype_is_real_integer(
    dtype: DTYPE_TYPE
) -> bool:
    return np.issubdtype(dtype, np.integer)

def dtype_is_real_floating(
    dtype: DTYPE_TYPE
) -> bool:
    return dtype == jax.dtypes.bfloat16 or np.issubdtype(dtype, np.floating)

def dtype_is_boolean(
    dtype: DTYPE_TYPE
) -> bool:
    return dtype == np.bool_ or dtype == bool

from .._common.implementations import *
if hasattr(jax.numpy, "__array_api_version__"):
    compat_module = jax.numpy
else:
    import jax.experimental.array_api as compat_module
abbreviate_array = get_abbreviate_array_function(
    backend=compat_module,
    default_integer_dtype=default_integer_dtype,
    func_dtype_is_real_floating=dtype_is_real_floating,
    func_dtype_is_real_integer=dtype_is_real_integer,
    func_dtype_is_boolean=dtype_is_boolean
)
def map_fn_over_arrays(
    data : Any, func : Callable[[ARRAY_TYPE], ARRAY_TYPE]
):
    return jax.tree.map(
        func,
        data
    )
pad_dim = get_pad_dim_function(
    backend=compat_module,
)


from typing import Protocol, TypeVar, Optional, Any, Tuple, Union, Type, TypedDict, Literal, Sequence, overload, Callable
from abc import abstractmethod
from array_api_typing.typing_compat._api_typing import ArrayAPINamespace as ArrayAPICompatNamespace, _NAMESPACE_ARRAY, _NAMESPACE_DEVICE, _NAMESPACE_DTYPE
from array_api_typing.typing_compat._array_typing import SetIndex, GetIndex
from ._at import AtResult

try:
    from array_api_extra._lib._at import Undef
except ImportError:
    from enum import Enum
    class Undef(Enum):
        UNDEF = 0

class ArrayAPINamespace(ArrayAPICompatNamespace[_NAMESPACE_ARRAY, _NAMESPACE_DEVICE, _NAMESPACE_DTYPE], Protocol[_NAMESPACE_ARRAY, _NAMESPACE_DEVICE, _NAMESPACE_DTYPE]):
    """
    From the array_api_extra
    https://github.com/data-apis/array-api-extra/blob/main/src/array_api_extra/_lib/_at.py
    """
    @abstractmethod
    def at(
        self,
        x: _NAMESPACE_ARRAY,
        idx : Union[SetIndex, Undef] = Undef.UNDEF,
        /
    ) -> "AtResult[_NAMESPACE_ARRAY]":
        """
        Returns an object that provides methods for elementwise operations
        on the array `x` using the `at` API.
        """
        pass

    """
    Utility Functions
    https://github.com/data-apis/array-api-extra/blob/main/src/array_api_extra/_lib/_funcs.py
    """
    @overload
    def apply_where(
        self,
        cond: _NAMESPACE_ARRAY,
        args: Union[_NAMESPACE_ARRAY, Tuple[_NAMESPACE_ARRAY, ...]],
        f1: Callable[..., _NAMESPACE_ARRAY],
        f2: Callable[..., _NAMESPACE_ARRAY],
        /,
    ) -> _NAMESPACE_ARRAY: 
        """
        Run one of two elementwise functions depending on a condition.

        Equivalent to ``f1(*args) if cond else fill_value`` performed elementwise
        when `fill_value` is defined, otherwise to ``f1(*args) if cond else f2(*args)``.

        Parameters
        ----------
        cond : array
            The condition, expressed as a boolean array.
        args : Array or tuple of Arrays
            Argument(s) to `f1` (and `f2`). Must be broadcastable with `cond`.
        f1 : callable
            Elementwise function of `args`, returning a single array.
            Where `cond` is True, output will be ``f1(arg0[cond], arg1[cond], ...)``.
        f2 : callable, optional
            Elementwise function of `args`, returning a single array.
            Where `cond` is False, output will be ``f2(arg0[cond], arg1[cond], ...)``.
            Mutually exclusive with `fill_value`.
        fill_value : Array or scalar, optional
            If provided, value with which to fill output array where `cond` is False.
            It does not need to be scalar; it needs however to be broadcastable with
            `cond` and `args`.
            Mutually exclusive with `f2`. You must provide one or the other.

        Returns
        -------
        Array
            An array with elements from the output of `f1` where `cond` is True and either
            the output of `f2` or `fill_value` where `cond` is False. The returned array has
            data type determined by type promotion rules between the output of `f1` and
            either `fill_value` or the output of `f2`.

        Notes
        -----
        ``xp.where(cond, f1(*args), f2(*args))`` requires explicitly evaluating `f1` even
        when `cond` is False, and `f2` when cond is True. This function evaluates each
        function only for their matching condition, if the backend allows for it.

        On Dask, `f1` and `f2` are applied to the individual chunks and should use functions
        from the namespace of the chunks.

        Examples
        --------
        >>> import array_api_strict as xp
        >>> import array_api_extra as xpx
        >>> a = xp.asarray([5, 4, 3])
        >>> b = xp.asarray([0, 2, 2])
        >>> def f(a, b):
        ...     return a // b
        >>> xpx.apply_where(b != 0, (a, b), f, fill_value=xp.nan)
        array([ nan,  2., 1.])
        """
    
    @overload
    def apply_where(
        self,
        cond: _NAMESPACE_ARRAY,
        args: Union[_NAMESPACE_ARRAY, Tuple[_NAMESPACE_ARRAY, ...]],
        f1: Callable[..., _NAMESPACE_ARRAY],
        f2: Callable[..., _NAMESPACE_ARRAY],
        /,
        *,
        fill_value: Optional[Union[_NAMESPACE_ARRAY, float, int, complex]] = None,
    ) -> _NAMESPACE_ARRAY: 
        pass

    @abstractmethod
    def atleast_nd(
        self,
        x: _NAMESPACE_ARRAY,
        /,
        *,
        ndim: int,
    ) -> _NAMESPACE_ARRAY:
        """
        Recursively expand the dimension of an array to at least `ndim`.

        Parameters
        ----------
        x : array
            Input array.
        ndim : int
            The minimum number of dimensions for the result.

        Returns
        -------
        array
            An array with ``res.ndim`` >= `ndim`.
            If ``x.ndim`` >= `ndim`, `x` is returned.
            If ``x.ndim`` < `ndim`, `x` is expanded by prepending new axes
            until ``res.ndim`` equals `ndim`.

        Examples
        --------
        >>> import array_api_strict as xp
        >>> import array_api_extra as xpx
        >>> x = xp.asarray([1])
        >>> xpx.atleast_nd(x, ndim=3, xp=xp)
        Array([[[1]]], dtype=array_api_strict.int64)

        >>> x = xp.asarray([[[1, 2],
        ...                  [3, 4]]])
        >>> xpx.atleast_nd(x, ndim=1, xp=xp) is x
        True
        """
        pass
    
    @abstractmethod
    def broadcast_shapes(
        self,
        *shapes : Tuple[Optional[float], ...]
    ) -> Tuple[Optional[float], ...]:
         """
        Compute the shape of the broadcasted arrays.

        Duplicates :func:`numpy.broadcast_shapes`, with additional support for
        None and NaN sizes.

        This is equivalent to ``xp.broadcast_arrays(arr1, arr2, ...)[0].shape``
        without needing to worry about the backend potentially deep copying
        the arrays.

        Parameters
        ----------
        *shapes : tuple[int | None, ...]
            Shapes of the arrays to broadcast.

        Returns
        -------
        tuple[int | None, ...]
            The shape of the broadcasted arrays.

        See Also
        --------
        numpy.broadcast_shapes : Equivalent NumPy function.
        array_api.broadcast_arrays : Function to broadcast actual arrays.

        Notes
        -----
        This function accepts the Array API's ``None`` for unknown sizes,
        as well as Dask's non-standard ``math.nan``.
        Regardless of input, the output always contains ``None`` for unknown sizes.

        Examples
        --------
        >>> import array_api_extra as xpx
        >>> xpx.broadcast_shapes((2, 3), (2, 1))
        (2, 3)
        >>> xpx.broadcast_shapes((4, 2, 3), (2, 1), (1, 3))
        (4, 2, 3)
        """
    
    def cov(
        self,
        m: _NAMESPACE_ARRAY,
        /,
    ) -> _NAMESPACE_ARRAY:
        """
        Estimate a covariance matrix.

        Covariance indicates the level to which two variables vary together.
        If we examine N-dimensional samples, :math:`X = [x_1, x_2, ... x_N]^T`,
        then the covariance matrix element :math:`C_{ij}` is the covariance of
        :math:`x_i` and :math:`x_j`. The element :math:`C_{ii}` is the variance
        of :math:`x_i`.

        This provides a subset of the functionality of ``numpy.cov``.

        Parameters
        ----------
        m : array
            A 1-D or 2-D array containing multiple variables and observations.
            Each row of `m` represents a variable, and each column a single
            observation of all those variables.
        xp : array_namespace, optional
            The standard-compatible namespace for `m`. Default: infer.

        Returns
        -------
        array
            The covariance matrix of the variables.

        Examples
        --------
        >>> import array_api_strict as xp
        >>> import array_api_extra as xpx

        Consider two variables, :math:`x_0` and :math:`x_1`, which
        correlate perfectly, but in opposite directions:

        >>> x = xp.asarray([[0, 2], [1, 1], [2, 0]]).T
        >>> x
        Array([[0, 1, 2],
            [2, 1, 0]], dtype=array_api_strict.int64)

        Note how :math:`x_0` increases while :math:`x_1` decreases. The covariance
        matrix shows this clearly:

        >>> xpx.cov(x, xp=xp)
        Array([[ 1., -1.],
            [-1.,  1.]], dtype=array_api_strict.float64)

        Note that element :math:`C_{0,1}`, which shows the correlation between
        :math:`x_0` and :math:`x_1`, is negative.

        Further, note how `x` and `y` are combined:

        >>> x = xp.asarray([-2.1, -1,  4.3])
        >>> y = xp.asarray([3,  1.1,  0.12])
        >>> X = xp.stack((x, y), axis=0)
        >>> xpx.cov(X, xp=xp)
        Array([[11.71      , -4.286     ],
            [-4.286     ,  2.14413333]], dtype=array_api_strict.float64)

        >>> xpx.cov(x, xp=xp)
        Array(11.71, dtype=array_api_strict.float64)

        >>> xpx.cov(y, xp=xp)
        Array(2.14413333, dtype=array_api_strict.float64)
        """
    
    def create_diagonal(
        self,
        x: _NAMESPACE_ARRAY,
        /,
        *,
        offset: int = 0,
    ) -> _NAMESPACE_ARRAY:
        """
        Construct a diagonal array.

        Parameters
        ----------
        x : array
            An array having shape ``(*batch_dims, k)``.
        offset : int, optional
            Offset from the leading diagonal (default is ``0``).
            Use positive ints for diagonals above the leading diagonal,
            and negative ints for diagonals below the leading diagonal.
        xp : array_namespace, optional
            The standard-compatible namespace for `x`. Default: infer.

        Returns
        -------
        array
            An array having shape ``(*batch_dims, k+abs(offset), k+abs(offset))`` with `x`
            on the diagonal (offset by `offset`).

        Examples
        --------
        >>> import array_api_strict as xp
        >>> import array_api_extra as xpx
        >>> x = xp.asarray([2, 4, 8])

        >>> xpx.create_diagonal(x, xp=xp)
        Array([[2, 0, 0],
            [0, 4, 0],
            [0, 0, 8]], dtype=array_api_strict.int64)

        >>> xpx.create_diagonal(x, offset=-2, xp=xp)
        Array([[0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0],
            [2, 0, 0, 0, 0],
            [0, 4, 0, 0, 0],
            [0, 0, 8, 0, 0]], dtype=array_api_strict.int64)
        """
    
    def kron(
        self,
        a: _NAMESPACE_ARRAY,
        b: _NAMESPACE_ARRAY,
        /,
    ) -> _NAMESPACE_ARRAY:
        """
        Kronecker product of two arrays.

        Computes the Kronecker product, a composite array made of blocks of the
        second array scaled by the first.

        Equivalent to ``numpy.kron`` for NumPy arrays.

        Parameters
        ----------
        a, b : Array | int | float | complex
            Input arrays or scalars. At least one must be an array.
        xp : array_namespace, optional
            The standard-compatible namespace for `a` and `b`. Default: infer.

        Returns
        -------
        array
            The Kronecker product of `a` and `b`.

        Notes
        -----
        The function assumes that the number of dimensions of `a` and `b`
        are the same, if necessary prepending the smallest with ones.
        If ``a.shape = (r0,r1,..,rN)`` and ``b.shape = (s0,s1,...,sN)``,
        the Kronecker product has shape ``(r0*s0, r1*s1, ..., rN*SN)``.
        The elements are products of elements from `a` and `b`, organized
        explicitly by::

            kron(a,b)[k0,k1,...,kN] = a[i0,i1,...,iN] * b[j0,j1,...,jN]

        where::

            kt = it * st + jt,  t = 0,...,N

        In the common 2-D case (N=1), the block structure can be visualized::

            [[ a[0,0]*b,   a[0,1]*b,  ... , a[0,-1]*b  ],
            [  ...                              ...   ],
            [ a[-1,0]*b,  a[-1,1]*b, ... , a[-1,-1]*b ]]

        Examples
        --------
        >>> import array_api_strict as xp
        >>> import array_api_extra as xpx
        >>> xpx.kron(xp.asarray([1, 10, 100]), xp.asarray([5, 6, 7]), xp=xp)
        Array([  5,   6,   7,  50,  60,  70, 500,
            600, 700], dtype=array_api_strict.int64)

        >>> xpx.kron(xp.asarray([5, 6, 7]), xp.asarray([1, 10, 100]), xp=xp)
        Array([  5,  50, 500,   6,  60, 600,   7,
                70, 700], dtype=array_api_strict.int64)

        >>> xpx.kron(xp.eye(2), xp.ones((2, 2)), xp=xp)
        Array([[1., 1., 0., 0.],
            [1., 1., 0., 0.],
            [0., 0., 1., 1.],
            [0., 0., 1., 1.]], dtype=array_api_strict.float64)

        >>> a = xp.reshape(xp.arange(100), (2, 5, 2, 5))
        >>> b = xp.reshape(xp.arange(24), (2, 3, 4))
        >>> c = xpx.kron(a, b, xp=xp)
        >>> c.shape
        (2, 10, 6, 20)
        >>> I = (1, 3, 0, 2)
        >>> J = (0, 2, 1)
        >>> J1 = (0,) + J             # extend to ndim=4
        >>> S1 = (1,) + b.shape
        >>> K = tuple(xp.asarray(I) * xp.asarray(S1) + xp.asarray(J1))
        >>> c[K] == a[I]*b[J]
        Array(True, dtype=array_api_strict.bool)
        """

    def nunique(
        self,
        x: _NAMESPACE_ARRAY,
        /,
    ) -> _NAMESPACE_ARRAY:
         """
        Count the number of unique elements in an array.

        Compatible with JAX and Dask, whose laziness would be otherwise
        problematic.

        Parameters
        ----------
        x : Array
            Input array.
        xp : array_namespace, optional
            The standard-compatible namespace for `x`. Default: infer.

        Returns
        -------
        array: 0-dimensional integer array
            The number of unique elements in `x`. It can be lazy.
        """
    
    def setdiff1d(
        self,
        x1: Union[_NAMESPACE_ARRAY, int, float, complex],
        x2: Union[_NAMESPACE_ARRAY, int, float, complex],
        /,
        *,
        assume_unique: bool = False,
    ) -> _NAMESPACE_ARRAY:
        """
        Find the set difference of two arrays.

        Return the unique values in `x1` that are not in `x2`.

        Parameters
        ----------
        x1 : array | int | float | complex | bool
            Input array.
        x2 : array
            Input comparison array.
        assume_unique : bool
            If ``True``, the input arrays are both assumed to be unique, which
            can speed up the calculation. Default is ``False``.
        xp : array_namespace, optional
            The standard-compatible namespace for `x1` and `x2`. Default: infer.

        Returns
        -------
        array
            1D array of values in `x1` that are not in `x2`. The result
            is sorted when `assume_unique` is ``False``, but otherwise only sorted
            if the input is sorted.

        Examples
        --------
        >>> import array_api_strict as xp
        >>> import array_api_extra as xpx

        >>> x1 = xp.asarray([1, 2, 3, 2, 4, 1])
        >>> x2 = xp.asarray([3, 4, 5, 6])
        >>> xpx.setdiff1d(x1, x2, xp=xp)
        Array([1, 2], dtype=array_api_strict.int64)
        """

    def sinc(
        self,
        x: _NAMESPACE_ARRAY,
        /,
    ) -> _NAMESPACE_ARRAY:
        r"""
        Return the normalized sinc function.

        The sinc function is equal to :math:`\sin(\pi x)/(\pi x)` for any argument
        :math:`x\ne 0`. ``sinc(0)`` takes the limit value 1, making ``sinc`` not
        only everywhere continuous but also infinitely differentiable.

        .. note::

            Note the normalization factor of ``pi`` used in the definition.
            This is the most commonly used definition in signal processing.
            Use ``sinc(x / xp.pi)`` to obtain the unnormalized sinc function
            :math:`\sin(x)/x` that is more common in mathematics.

        Parameters
        ----------
        x : array
            Array (possibly multi-dimensional) of values for which to calculate
            ``sinc(x)``. Must have a real floating point dtype.
        xp : array_namespace, optional
            The standard-compatible namespace for `x`. Default: infer.

        Returns
        -------
        array
            ``sinc(x)`` calculated elementwise, which has the same shape as the input.

        Notes
        -----
        The name sinc is short for "sine cardinal" or "sinus cardinalis".

        The sinc function is used in various signal processing applications,
        including in anti-aliasing, in the construction of a Lanczos resampling
        filter, and in interpolation.

        For bandlimited interpolation of discrete-time signals, the ideal
        interpolation kernel is proportional to the sinc function.

        References
        ----------
        #. Weisstein, Eric W. "Sinc Function." From MathWorld--A Wolfram Web
        Resource. https://mathworld.wolfram.com/SincFunction.html
        #. Wikipedia, "Sinc function",
        https://en.wikipedia.org/wiki/Sinc_function

        Examples
        --------
        >>> import array_api_strict as xp
        >>> import array_api_extra as xpx
        >>> x = xp.linspace(-4, 4, 41)
        >>> xpx.sinc(x, xp=xp)
        Array([-3.89817183e-17, -4.92362781e-02,
            -8.40918587e-02, -8.90384387e-02,
            -5.84680802e-02,  3.89817183e-17,
                6.68206631e-02,  1.16434881e-01,
                1.26137788e-01,  8.50444803e-02,
            -3.89817183e-17, -1.03943254e-01,
            -1.89206682e-01, -2.16236208e-01,
            -1.55914881e-01,  3.89817183e-17,
                2.33872321e-01,  5.04551152e-01,
                7.56826729e-01,  9.35489284e-01,
                1.00000000e+00,  9.35489284e-01,
                7.56826729e-01,  5.04551152e-01,
                2.33872321e-01,  3.89817183e-17,
            -1.55914881e-01, -2.16236208e-01,
            -1.89206682e-01, -1.03943254e-01,
            -3.89817183e-17,  8.50444803e-02,
                1.26137788e-01,  1.16434881e-01,
                6.68206631e-02,  3.89817183e-17,
            -5.84680802e-02, -8.90384387e-02,
            -8.40918587e-02, -4.92362781e-02,
            -3.89817183e-17], dtype=array_api_strict.float64)
        """

    """
    Delegation functions
    https://github.com/data-apis/array-api-extra/blob/main/src/array_api_extra/_delegation.py
    """
    def isclose(
        a: Union[_NAMESPACE_ARRAY, int, float, complex],
        b: Union[_NAMESPACE_ARRAY, int, float, complex],
        *,
        rtol: float = 1e-05,
        atol: float = 1e-08,
        equal_nan: bool = False,
    ) -> _NAMESPACE_ARRAY:
        """
        Return a boolean array where two arrays are element-wise equal within a tolerance.

        The tolerance values are positive, typically very small numbers. The relative
        difference ``(rtol * abs(b))`` and the absolute difference `atol` are added together
        to compare against the absolute difference between `a` and `b`.

        NaNs are treated as equal if they are in the same place and if ``equal_nan=True``.
        Infs are treated as equal if they are in the same place and of the same sign in both
        arrays.

        Parameters
        ----------
        a, b : Array | int | float | complex | bool
            Input objects to compare. At least one must be an array.
        rtol : array_like, optional
            The relative tolerance parameter (see Notes).
        atol : array_like, optional
            The absolute tolerance parameter (see Notes).
        equal_nan : bool, optional
            Whether to compare NaN's as equal. If True, NaN's in `a` will be considered
            equal to NaN's in `b` in the output array.
        xp : array_namespace, optional
            The standard-compatible namespace for `a` and `b`. Default: infer.

        Returns
        -------
        Array
            A boolean array of shape broadcasted from `a` and `b`, containing ``True`` where
            `a` is close to `b`, and ``False`` otherwise.

        Warnings
        --------
        The default `atol` is not appropriate for comparing numbers with magnitudes much
        smaller than one (see notes).

        See Also
        --------
        math.isclose : Similar function in stdlib for Python scalars.

        Notes
        -----
        For finite values, `isclose` uses the following equation to test whether two
        floating point values are equivalent::

            absolute(a - b) <= (atol + rtol * absolute(b))

        Unlike the built-in `math.isclose`,
        the above equation is not symmetric in `a` and `b`,
        so that ``isclose(a, b)`` might be different from ``isclose(b, a)`` in some rare
        cases.

        The default value of `atol` is not appropriate when the reference value `b` has
        magnitude smaller than one. For example, it is unlikely that ``a = 1e-9`` and
        ``b = 2e-9`` should be considered "close", yet ``isclose(1e-9, 2e-9)`` is ``True``
        with default settings. Be sure to select `atol` for the use case at hand, especially
        for defining the threshold below which a non-zero value in `a` will be considered
        "close" to a very small or zero value in `b`.

        The comparison of `a` and `b` uses standard broadcasting, which means that `a` and
        `b` need not have the same shape in order for ``isclose(a, b)`` to evaluate to
        ``True``.

        `isclose` is not defined for non-numeric data types.
        ``bool`` is considered a numeric data-type for this purpose.
        """
    
    def pad(
        self,
        x: _NAMESPACE_ARRAY,
        pad_width: Union[int, Tuple[int, int], Sequence[Tuple[int, int]]],
        mode: Literal["constant"] = "constant",
        *,
        constant_values: Union[int, float, complex] = 0,
    ) -> _NAMESPACE_ARRAY:
        """
        Pad the input array.

        Parameters
        ----------
        x : array
            Input array.
        pad_width : int or tuple of ints or sequence of pairs of ints
            Pad the input array with this many elements from each side.
            If a sequence of tuples, ``[(before_0, after_0), ... (before_N, after_N)]``,
            each pair applies to the corresponding axis of ``x``.
            A single tuple, ``(before, after)``, is equivalent to a list of ``x.ndim``
            copies of this tuple.
        mode : str, optional
            Only "constant" mode is currently supported, which pads with
            the value passed to `constant_values`.
        constant_values : python scalar, optional
            Use this value to pad the input. Default is zero.
        xp : array_namespace, optional
            The standard-compatible namespace for `x`. Default: infer.

        Returns
        -------
        array
            The input array,
            padded with ``pad_width`` elements equal to ``constant_values``.
        """


from typing import Union, Optional, Tuple, Any
from ._typing import ARRAY_TYPE, DTYPE_TYPE, DEVICE_TYPE, RNG_TYPE
import numpy as np

__all__ = [
    "random_number_generator",
    "random_discrete_uniform",
    "random_uniform",
    "random_exponential",
    "random_normal",
    "random_geometric",
    "random_permutation"
]

def random_number_generator(
    seed : Optional[int] = None,
    *,
    device : Optional[DEVICE_TYPE] = None
) -> RNG_TYPE:
    return np.random.default_rng(seed)

def random_discrete_uniform(
    shape : Union[int, Tuple[int, ...]], 
    /,
    from_num : int, 
    to_num : int, 
    *,
    rng : RNG_TYPE, 
    dtype : Optional[DTYPE_TYPE] = None, 
    device : Optional[DEVICE_TYPE] = None
) -> Tuple[RNG_TYPE, ARRAY_TYPE]:
    t = rng.integers(int(from_num), int(to_num), size=shape)
    if dtype is not None:
        t = t.astype(dtype)
    return rng, t

def random_uniform(
    shape: Union[int, Tuple[int, ...]], 
    /,
    *,
    rng : RNG_TYPE, 
    low : float = 0.0, high : float = 1.0,
    dtype : Optional[DTYPE_TYPE] = None, 
    device : Optional[DEVICE_TYPE] = None
) -> Tuple[RNG_TYPE, ARRAY_TYPE]:
    t = rng.uniform(float(low), float(high), size=shape)
    if dtype is not None:
        t = t.astype(dtype)
    return rng, t

def random_exponential(
    shape: Union[int, Tuple[int, ...]], 
    /,
    *,
    rng : RNG_TYPE,  
    lambd : float = 1.0, 
    dtype : Optional[DTYPE_TYPE] = None, 
    device : Optional[DEVICE_TYPE] = None
) -> Tuple[RNG_TYPE, ARRAY_TYPE]:
    t = rng.exponential(1.0 / float(lambd), size=shape)
    if dtype is not None:
        t = t.astype(dtype)
    return rng, t

def random_normal(
    shape: Union[int, Tuple[int, ...]],
    /,
    *, 
    rng : RNG_TYPE,
    mean : float = 0.0, std : float = 1.0, 
    dtype : Optional[DTYPE_TYPE] = None, 
    device : Optional[Any] = None
) -> Tuple[RNG_TYPE, ARRAY_TYPE]:
    t = rng.normal(mean, std, size=shape)
    if dtype is not None:
        t = t.astype(dtype)
    return rng, t


def random_geometric(
    shape: Union[int, Tuple[int, ...]], 
    /,
    *,
    p: float, 
    rng: RNG_TYPE,
    dtype: Optional[DTYPE_TYPE] = None, 
    device: Optional[Any] = None
) -> Tuple[RNG_TYPE, ARRAY_TYPE]:
    t = rng.geometric(p, size=shape)
    if dtype is not None:
        t = t.astype(dtype)
    return rng, t

def random_permutation(
    n : int,
    /,
    *,
    rng: RNG_TYPE,
    dtype: Optional[DTYPE_TYPE] = None,
    device: Optional[DEVICE_TYPE] = None
) -> Tuple[RNG_TYPE, ARRAY_TYPE]:
    t = rng.permutation(n)
    if dtype is not None:
        t = t.astype(dtype)
    return rng, t


import torch

__all__ = [
    'ARRAY_TYPE',
    'DTYPE_TYPE',
    'DEVICE_TYPE',
    'RNG_TYPE',
]

ARRAY_TYPE = torch.Tensor
DTYPE_TYPE = torch.dtype
DEVICE_TYPE = torch.device
RNG_TYPE = torch.Generator